"""Shared auth setup logic for onboarding and CLI."""

import os

from sqlsaber.application.prompts import Prompter
from sqlsaber.config import providers
from sqlsaber.config.api_keys import APIKeyManager
from sqlsaber.config.auth import AuthConfigManager, AuthMethod
from sqlsaber.theme.manager import create_console

console = create_console()


async def select_provider(prompter: Prompter, default: str = "anthropic") -> str | None:
    """Interactive provider selection.

    Args:
        prompter: Prompter instance for interaction
        default: Default provider to select

    Returns:
        Selected provider name or None if cancelled
    """
    provider = await prompter.select(
        "Select AI provider:", choices=providers.all_keys(), default=default
    )
    return provider


async def configure_api_key(
    provider: str, api_key_manager: APIKeyManager, auth_manager: AuthConfigManager
) -> bool:
    """Configure API key for a provider.

    Args:
        provider: Provider name
        api_key_manager: APIKeyManager instance
        auth_manager: AuthConfigManager instance

    Returns:
        True if API key configured successfully, False otherwise
    """
    # Get API key (cascades env -> keyring -> prompt)
    api_key = api_key_manager.get_api_key(provider)

    if api_key:
        auth_manager.set_auth_method(AuthMethod.API_KEY)
        return True

    return False


async def setup_auth(
    prompter: Prompter,
    auth_manager: AuthConfigManager,
    api_key_manager: APIKeyManager,
    default_provider: str = "anthropic",
) -> tuple[bool, str | None]:
    """Interactive authentication setup.

    Args:
        prompter: Prompter instance for interaction
        auth_manager: AuthConfigManager instance
        api_key_manager: APIKeyManager instance
        default_provider: Default provider to select

    Returns:
        Tuple of (success: bool, provider: str | None)
    """
    provider = await select_provider(prompter, default=default_provider)

    if provider is None:
        return False, None

    env_var = api_key_manager.get_env_var_name(provider)
    api_key_in_env = bool(os.getenv(env_var))
    api_key_in_keyring = api_key_manager.has_stored_api_key(provider)

    if api_key_in_env or api_key_in_keyring:
        parts: list[str] = []
        if api_key_in_keyring:
            parts.append("stored API key")
        if api_key_in_env:
            parts.append(f"{env_var} environment variable")
        summary = ", ".join(parts)
        console.print(
            f"[info]Existing authentication found for {provider}: {summary}[/info]"
        )

    # API key flow
    if api_key_in_keyring:
        reset_api_key = await prompter.confirm(
            f"{provider.title()} API key is stored in your keyring. Reset before continuing?",
            default=False,
        )
        if not reset_api_key:
            console.print(
                "[warning]No changes made to stored API key credentials.[/warning]"
            )
            return True, None
        if not api_key_manager.delete_api_key(provider):
            console.print(
                "[error]Failed to remove existing API key credentials.[/error]"
            )
            return False, None
        console.print(
            f"[muted]{provider.title()} API key removed from keyring.[/muted]"
        )
        api_key_in_keyring = False

    if api_key_in_env:
        console.print(
            f"[muted]{env_var} is set in your environment. Update it there if you need a new value.[/muted]"
        )

    console.print()
    console.print(f"[dim]To use {provider.title()}, you need an API key.[/dim]")
    console.print(f"[dim]You can set the {env_var} environment variable,[/dim]")
    console.print("[dim]or enter it now to store securely in your OS keychain.[/dim]")
    console.print()

    api_key_configured = await configure_api_key(
        provider, api_key_manager, auth_manager
    )

    if api_key_configured:
        console.print(
            f"[green]✓ {provider.title()} API key configured successfully![/green]"
        )
        return True, provider

    console.print("[warning]No API key provided.[/warning]")
    return False, None



"""SQL query validation and security using sqlglot AST analysis."""

from collections.abc import Callable
from dataclasses import dataclass

import sqlglot
from sqlglot import exp
from sqlglot.errors import ParseError

# DML/DDL operations that can be unlocked in "dangerous" mode
WRITE_DML_DDL_NODES: set[type[exp.Expression]] = {
    # DML operations
    exp.Insert,
    exp.Update,
    exp.Delete,
    exp.Merge,
    # MySQL specific
    exp.Replace,
    # DDL operations (non-destructive)
    exp.Create,
    exp.Alter,
    exp.AlterRename,
}

# Operations that are always prohibited, regardless of mode
ALWAYS_BLOCKED_NODES: set[type[exp.Expression]] = {
    # Transaction control
    exp.Transaction,
    # Analysis and maintenance
    exp.Analyze,
    # Data loading/copying
    exp.Copy,
    exp.LoadData,
    # Session and configuration
    exp.Set,
    exp.Use,
    exp.Pragma,
    # Security
    exp.Grant,
    exp.Revoke,
    # Database operations
    exp.Attach,
    exp.Detach,
    # Locking and process control
    exp.Lock,
    exp.Kill,
    # Commands
    exp.Command,
    # Destructive schema/data operations (no safeguards possible)
    exp.Drop,
    exp.TruncateTable,
}

try:
    vacuum_type = getattr(exp, "Vacuum", None)
    if vacuum_type is not None:
        ALWAYS_BLOCKED_NODES.add(vacuum_type)
except AttributeError:
    pass

# Dangerous functions by dialect that can read files or execute commands
DANGEROUS_FUNCTIONS_BY_DIALECT: dict[str, set[str]] = {
    "postgres": {
        "pg_read_file",
        "pg_read_binary_file",
        "pg_ls_dir",
        "pg_stat_file",
        "pg_logdir_ls",
        "dblink",
        "dblink_exec",
    },
    "mysql": {
        "load_file",
        "sys_eval",
        "sys_exec",
    },
    "sqlite": {
        "readfile",
        "writefile",
    },
    "tsql": {
        "xp_cmdshell",
    },
}


@dataclass
class GuardResult:
    """Result of SQL query validation."""

    allowed: bool
    reason: str | None = None
    is_select: bool = False
    query_type: str | None = None  # "select" | "dml" | "ddl" | "other"


def is_select_like(stmt: exp.Expression) -> bool:
    """Check if statement is a SELECT-like query.

    Handles CTEs (WITH) and set operations (UNION/INTERSECT/EXCEPT).
    """
    root = stmt
    # WITH wraps another statement
    if isinstance(root, exp.With):
        root = root.this
    return isinstance(root, (exp.Select, exp.Union, exp.Except, exp.Intersect))


def classify_statement(stmt: exp.Expression) -> str:
    """Classify statement as select/dml/ddl/other.

    Returns:
        "select" for SELECT-like queries
        "dml" for INSERT/UPDATE/DELETE/MERGE/REPLACE
        "ddl" for CREATE/DROP/ALTER/TRUNCATE
        "other" for anything else
    """
    if is_select_like(stmt):
        return "select"

    root = stmt
    if isinstance(root, exp.With):
        root = root.this

    if isinstance(root, (exp.Insert, exp.Update, exp.Delete, exp.Merge, exp.Replace)):
        return "dml"

    if isinstance(
        root,
        (exp.Create, exp.Alter, exp.AlterRename),
    ):
        return "ddl"

    # DROP and TRUNCATE are blocked, but classify them for error messages
    if isinstance(root, (exp.Drop, exp.TruncateTable)):
        return "ddl"

    return "other"


def has_unfiltered_mutation(stmt: exp.Expression) -> str | None:
    """Check for UPDATE/DELETE without WHERE clause.

    These operations are dangerous because they affect all rows in a table.
    """
    for node in stmt.walk():
        if isinstance(node, exp.Update):
            if not node.args.get("where"):
                return (
                    "UPDATE without WHERE clause is not allowed (would affect all rows)"
                )
        if isinstance(node, exp.Delete):
            if not node.args.get("where"):
                return (
                    "DELETE without WHERE clause is not allowed (would affect all rows)"
                )
    return None


def has_prohibited_nodes(
    stmt: exp.Expression, allow_dangerous: bool = False
) -> str | None:
    """Walk AST to find any prohibited operations.

    In read-only mode (allow_dangerous=False):
      - Block DML/DDL (WRITE_DML_DDL_NODES)
      - Block always-blocked operations (ALWAYS_BLOCKED_NODES)
      - Block SELECT INTO
      - Block locking clauses (FOR UPDATE/FOR SHARE)

    In dangerous mode (allow_dangerous=True):
      - Allow DML/DDL
      - Still block ALWAYS_BLOCKED_NODES, SELECT INTO, locking clauses
      - Block UPDATE/DELETE without WHERE clause
    """
    for node in stmt.walk():
        # Operations that are never allowed
        if isinstance(node, tuple(ALWAYS_BLOCKED_NODES)):
            return f"Prohibited operation: {type(node).__name__}"

        # DML/DDL writes are only allowed in dangerous mode
        if not allow_dangerous and isinstance(node, tuple(WRITE_DML_DDL_NODES)):
            return f"Prohibited operation: {type(node).__name__}"

        # Block SELECT INTO (Postgres-style table creation)
        if isinstance(node, exp.Select) and node.args.get("into"):
            return "SELECT INTO is not allowed"

        # Block locking clauses (FOR UPDATE/FOR SHARE)
        if isinstance(node, exp.Select):
            locks = node.args.get("locks")
            if locks:
                return "SELECT with locking clause (FOR UPDATE/SHARE) is not allowed"

    # In dangerous mode, block unfiltered mutations
    if allow_dangerous:
        reason = has_unfiltered_mutation(stmt)
        if reason:
            return reason

    return None


def has_dangerous_functions(stmt: exp.Expression, dialect: str) -> str | None:
    """Check for dangerous functions that can read files or execute commands."""
    deny_set = DANGEROUS_FUNCTIONS_BY_DIALECT.get(dialect)
    if not deny_set:
        return None

    deny_lower = {f.lower() for f in deny_set}

    for fn in stmt.find_all(exp.Func):
        name = fn.name
        if name and name.lower() in deny_lower:
            return f"Use of dangerous function '{name}' is not allowed"

    return None


def validate_read_only(sql: str, dialect: str = "ansi") -> GuardResult:
    """Validate that SQL query is read-only using AST analysis.

    Args:
        sql: SQL query to validate
        dialect: SQL dialect (postgres, mysql, sqlite, tsql, etc.)

    Returns:
        GuardResult with validation outcome
    """
    try:
        statements = sqlglot.parse(sql, read=dialect)
    except ParseError as e:
        return GuardResult(False, f"Unable to parse query safely: {e}")
    except Exception as e:
        return GuardResult(False, f"Error parsing query: {e}")

    # Only allow single statements
    if len(statements) != 1:
        return GuardResult(
            False,
            f"Only single SELECT statements are allowed (got {len(statements)} statements)",
        )

    stmt = statements[0]
    if stmt is None:
        return GuardResult(False, "Unable to parse query - empty statement")

    # Must be a SELECT-like statement
    if not is_select_like(stmt):
        return GuardResult(False, "Only SELECT-like statements are allowed")

    # Check for prohibited operations in the AST
    reason = has_prohibited_nodes(stmt)
    if reason:
        return GuardResult(False, reason)

    # Check for dangerous functions
    reason = has_dangerous_functions(stmt, dialect)
    if reason:
        return GuardResult(False, reason)

    return GuardResult(True, None, is_select=True, query_type="select")


def validate_sql(
    sql: str, dialect: str = "ansi", allow_dangerous: bool = False
) -> GuardResult:
    """Validate SQL with optional write/DDL allowance.

    In read-only mode (default): same behavior as validate_read_only.
    In dangerous mode: allow DML/DDL, but still enforce:
      - single statement
      - parseability
      - no dangerous functions (file IO, command exec, etc.)

    Args:
        sql: SQL query to validate
        dialect: SQL dialect (postgres, mysql, sqlite, tsql, etc.)
        allow_dangerous: If True, allow DML/DDL statements

    Returns:
        GuardResult with validation outcome
    """
    if not allow_dangerous:
        return validate_read_only(sql, dialect)

    try:
        statements = sqlglot.parse(sql, read=dialect)
    except ParseError as e:
        return GuardResult(False, f"Unable to parse query safely: {e}")
    except Exception as e:
        return GuardResult(False, f"Error parsing query: {e}")

    if len(statements) != 1:
        return GuardResult(
            False,
            f"Only single statements are allowed (got {len(statements)} statements)",
        )

    stmt = statements[0]
    if stmt is None:
        return GuardResult(False, "Unable to parse query - empty statement")

    # Still enforce function-level sandbox even in dangerous mode
    reason = has_dangerous_functions(stmt, dialect)
    if reason:
        return GuardResult(False, reason)

    # Still enforce always-blocked operations (COPY, LOAD DATA, SET, PRAGMA, etc.)
    reason = has_prohibited_nodes(stmt, allow_dangerous=True)
    if reason:
        return GuardResult(False, reason)

    query_type = classify_statement(stmt)
    return GuardResult(
        True, None, is_select=(query_type == "select"), query_type=query_type
    )


def add_limit(sql: str, dialect: str = "ansi", limit: int = 100) -> str:
    """Add LIMIT clause to query if not already present.

    Args:
        sql: SQL query
        dialect: SQL dialect for proper rendering
        limit: Maximum number of rows to return

    Returns:
        SQL with LIMIT clause added (or original if LIMIT already exists)
    """
    # Strip trailing semicolon to ensure clean parsing and modification
    # This handles cases where models generate SQL with a trailing semicolon
    sql = sql.strip().rstrip(";")

    try:
        statements = sqlglot.parse(sql, read=dialect)
        if len(statements) != 1:
            return sql

        stmt = statements[0]
        if stmt is None:
            return sql

        # Check if LIMIT/FETCH already exists
        has_limit = any(isinstance(n, (exp.Limit, exp.Fetch)) for n in stmt.walk())
        if has_limit:
            return stmt.sql(dialect=dialect)

        # Add LIMIT - sqlglot will render appropriately for dialect
        # (LIMIT for most, TOP for SQL Server, FETCH FIRST for Oracle)
        limit_method: Callable[[int], exp.Expression] | None = getattr(
            stmt, "limit", None
        )
        if limit_method is not None:
            limited_stmt = limit_method(limit)
            return limited_stmt.sql(dialect=dialect)
        return stmt.sql(dialect=dialect)

    except Exception:
        # If parsing/transformation fails, fall back to simple string append
        # This maintains backward compatibility
        sql_upper = sql.strip().upper()
        if "LIMIT" not in sql_upper:
            return f"{sql.rstrip(';')} LIMIT {limit};"
        return sql



from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

from rich.console import Console

from sqlsaber.cli.display import DisplayManager

if TYPE_CHECKING:
    from sqlsaber.agents.pydantic_ai_agent import SQLSaberAgent
    from sqlsaber.cli.usage import SessionUsage
    from sqlsaber.threads.manager import ThreadManager


@dataclass
class CommandContext:
    """Context passed to slash command handlers."""

    console: Console
    agent: "SQLSaberAgent"
    thread_manager: "ThreadManager"
    on_clear_history: Callable[[], None]
    session_usage: "SessionUsage | None" = None


@dataclass
class CommandResult:
    """Result of command processing."""

    handled: bool
    should_exit: bool = False


class SlashCommandProcessor:
    """Processes slash commands and special inputs."""

    EXIT_COMMANDS = {"/exit", "/quit", "exit", "quit"}

    async def process(self, user_query: str, context: CommandContext) -> CommandResult:
        """
        Process a user query to see if it's a command.
        Returns CommandResult indicating if it was handled and if we should exit.
        """
        query = user_query.strip().lower()

        # Handle exit commands
        if query in self.EXIT_COMMANDS or any(
            query.startswith(cmd) for cmd in self.EXIT_COMMANDS
        ):
            return await self._handle_exit(context)

        if query == "/clear":
            return await self._handle_clear(context)

        if query == "/thinking on":
            return await self._handle_thinking(context, enabled=True)

        if query == "/thinking off":
            return await self._handle_thinking(context, enabled=False)

        return CommandResult(handled=False)

    async def _handle_exit(self, context: CommandContext) -> CommandResult:
        """Handle exit commands."""
        if context.session_usage is not None:
            display = DisplayManager(context.console)
            display.show_session_summary(context.session_usage)
        ended_thread_id = await context.thread_manager.end_current_thread()
        if ended_thread_id:
            hint = f"saber threads resume {ended_thread_id}"
            context.console.print(
                f"[muted]You can continue this thread using:[/muted] {hint}"
            )
        return CommandResult(handled=True, should_exit=True)

    async def _handle_clear(self, context: CommandContext) -> CommandResult:
        """Handle /clear command."""
        context.on_clear_history()
        await context.thread_manager.clear_current_thread()
        context.console.print("[success]Conversation history cleared.[/success]\n")
        return CommandResult(handled=True)

    async def _handle_thinking(
        self, context: CommandContext, enabled: bool
    ) -> CommandResult:
        """Handle /thinking on/off commands."""
        context.agent.set_thinking(enabled=enabled)
        state = "enabled" if enabled else "disabled"
        context.console.print(f"[success]✓ Thinking {state}[/success]\n")
        return CommandResult(handled=True)



from unittest.mock import AsyncMock, MagicMock

import pytest

from sqlsaber.cli.slash_commands import CommandContext, SlashCommandProcessor


@pytest.fixture
def mock_context():
    """Fixture for a mocked CommandContext."""
    return CommandContext(
        console=MagicMock(),
        agent=MagicMock(),
        thread_manager=AsyncMock(),
        on_clear_history=MagicMock(),
    )


@pytest.fixture
def processor():
    """Fixture for SlashCommandProcessor."""
    return SlashCommandProcessor()


@pytest.mark.asyncio
async def test_process_unknown_command(processor, mock_context):
    """Test processing an unknown command."""
    result = await processor.process("hello world", mock_context)
    assert result.handled is False
    assert result.should_exit is False


@pytest.mark.asyncio
async def test_process_exit_command(processor, mock_context):
    """Test processing exit commands."""
    # Setup thread manager to return a thread ID (simulating an active thread ending)
    mock_context.thread_manager.end_current_thread.return_value = "thread-123"

    for cmd in ["/exit", "/quit", "exit", "quit", "QUIT", "EXIT", "/EXIT", "/QUIT"]:
        result = await processor.process(cmd, mock_context)

        assert result.handled is True
        assert result.should_exit is True
        mock_context.thread_manager.end_current_thread.assert_called()

        # Verify hint is printed
        mock_context.console.print.assert_called()
        args, _ = mock_context.console.print.call_args
        assert "saber threads resume thread-123" in args[0]


@pytest.mark.asyncio
async def test_process_clear_command(processor, mock_context):
    """Test processing /clear command."""
    result = await processor.process("/clear", mock_context)

    assert result.handled is True
    assert result.should_exit is False

    # Verify actions
    mock_context.on_clear_history.assert_called_once()
    mock_context.thread_manager.clear_current_thread.assert_called_once()
    mock_context.console.print.assert_called()


@pytest.mark.asyncio
async def test_process_thinking_on(processor, mock_context):
    """Test processing /thinking on command."""
    result = await processor.process("/thinking on", mock_context)

    assert result.handled is True
    assert result.should_exit is False

    mock_context.agent.set_thinking.assert_called_once_with(enabled=True)
    mock_context.console.print.assert_called()


@pytest.mark.asyncio
async def test_process_thinking_off(processor, mock_context):
    """Test processing /thinking off command."""
    result = await processor.process("/thinking off", mock_context)

    assert result.handled is True
    assert result.should_exit is False

    mock_context.agent.set_thinking.assert_called_once_with(enabled=False)
    mock_context.console.print.assert_called()



"""SQLite storage for pydantic-ai thread snapshots.

Each thread represents a session (interactive or non-interactive) and stores the
complete pydantic-ai message history as a snapshot. On every completed run in the
same session, we overwrite the snapshot with the new full history.

This design intentionally avoids per-run append logs and mirrors pydantic-ai's
recommended approach of serializing ModelMessage[] with ModelMessagesTypeAdapter.
"""

import asyncio
import time
import uuid
from pathlib import Path
from typing import Any

import aiosqlite
import platformdirs
from pydantic_ai.messages import ModelMessage, ModelMessagesTypeAdapter

from sqlsaber.config.logging import get_logger

logger = get_logger(__name__)


SCHEMA_SQL = """
CREATE TABLE IF NOT EXISTS threads (
    id TEXT PRIMARY KEY,
    database_name TEXT,
    title TEXT,
    created_at REAL NOT NULL,
    ended_at REAL,
    last_activity_at REAL NOT NULL,
    model_name TEXT,
    messages_json BLOB NOT NULL,
    extra_metadata TEXT
);

CREATE INDEX IF NOT EXISTS idx_threads_dbname ON threads(database_name);
CREATE INDEX IF NOT EXISTS idx_threads_activity ON threads(last_activity_at);
"""


class Thread:
    """Thread metadata."""

    def __init__(
        self,
        id: str,
        database_name: str | None,
        title: str | None,
        created_at: float,
        ended_at: float | None,
        last_activity_at: float,
        model_name: str | None,
    ) -> None:
        self.id = id
        self.database_name = database_name
        self.title = title
        self.created_at = created_at
        self.ended_at = ended_at
        self.last_activity_at = last_activity_at
        self.model_name = model_name


class ThreadStorage:
    """Handles SQLite storage for pydantic-ai thread snapshots."""

    def __init__(self) -> None:
        self.db_path = Path(platformdirs.user_config_dir("sqlsaber")) / "threads.db"
        self._lock = asyncio.Lock()
        self._initialized = False

    async def _init_db(self) -> None:
        if self._initialized:
            return
        try:
            self.db_path.parent.mkdir(parents=True, exist_ok=True)
            async with aiosqlite.connect(self.db_path) as db:
                await db.executescript(SCHEMA_SQL)
                await db.commit()
            self._initialized = True
            logger.info("threads.db.init", path=str(self.db_path))
        except Exception as e:  # pragma: no cover - best-effort persistence
            logger.warning("threads.db.init_failed", error=str(e))

    async def save_snapshot(
        self,
        *,
        messages_json: bytes,
        database_name: str | None,
        thread_id: str | None = None,
        extra_metadata: str | None = None,
    ) -> str:
        """Create or update a thread snapshot."""
        await self._init_db()
        now = time.time()

        if thread_id is None:
            thread_id = str(uuid.uuid4())
            try:
                async with self._lock, aiosqlite.connect(self.db_path) as db:
                    await db.execute(
                        """
                        INSERT INTO threads (
                            id, database_name, created_at,
                            last_activity_at, messages_json, extra_metadata
                        ) VALUES (?, ?, ?, ?, ?, ?)
                        """,
                        (
                            thread_id,
                            database_name,
                            now,
                            now,
                            messages_json,
                            extra_metadata,
                        ),
                    )
                    await db.commit()
                logger.info("threads.create", thread_id=thread_id)
                return thread_id
            except Exception as e:  # pragma: no cover
                logger.warning("threads.create_failed", error=str(e))
                return thread_id
        else:
            try:
                async with self._lock, aiosqlite.connect(self.db_path) as db:
                    await db.execute(
                        """
                        UPDATE threads
                        SET last_activity_at = ?,
                            messages_json = ?,
                            extra_metadata = COALESCE(?, extra_metadata)
                        WHERE id = ?
                        """,
                        (
                            now,
                            messages_json,
                            extra_metadata,
                            thread_id,
                        ),
                    )
                    await db.commit()
                logger.info("threads.update_snapshot", thread_id=thread_id)
                return thread_id
            except Exception as e:  # pragma: no cover
                logger.warning(
                    "threads.update_snapshot_failed", thread_id=thread_id, error=str(e)
                )
                return thread_id

    async def save_metadata(
        self,
        *,
        thread_id: str,
        title: str | None = None,
        model_name: str | None = None,
    ) -> bool:
        """Update thread metadata (title/model/extra). Only provided fields are updated."""
        await self._init_db()

        try:
            async with self._lock, aiosqlite.connect(self.db_path) as db:
                await db.execute(
                    """
                    UPDATE threads
                    SET title = ?, model_name = ?
                    WHERE id = ?
                    """,
                    (title, model_name, thread_id),
                )
                await db.commit()
            logger.info("threads.update_metadata", thread_id=thread_id)
            return True
        except Exception as e:  # pragma: no cover
            logger.warning(
                "threads.update_metadata_failed", thread_id=thread_id, error=str(e)
            )
            return False

    async def end_thread(self, thread_id: str) -> bool:
        await self._init_db()
        try:
            async with self._lock, aiosqlite.connect(self.db_path) as db:
                await db.execute(
                    "UPDATE threads SET ended_at = ?, last_activity_at = ? WHERE id = ?",
                    (time.time(), time.time(), thread_id),
                )
                await db.commit()
            logger.info("threads.end", thread_id=thread_id)
            return True
        except Exception as e:  # pragma: no cover
            logger.warning("threads.end_failed", thread_id=thread_id, error=str(e))
            return False

    async def get_thread(self, thread_id: str) -> Thread | None:
        await self._init_db()
        try:
            async with aiosqlite.connect(self.db_path) as db:
                async with db.execute(
                    """
                    SELECT id, database_name, title, created_at, ended_at,
                           last_activity_at, model_name
                    FROM threads WHERE id = ?
                    """,
                    (thread_id,),
                ) as cur:
                    row = await cur.fetchone()
                    if not row:
                        return None
                    return Thread(
                        id=row[0],
                        database_name=row[1],
                        title=row[2],
                        created_at=row[3],
                        ended_at=row[4],
                        last_activity_at=row[5],
                        model_name=row[6],
                    )
        except Exception as e:  # pragma: no cover
            logger.warning("threads.get_failed", thread_id=thread_id, error=str(e))
            return None

    async def get_thread_messages(self, thread_id: str) -> list[ModelMessage]:
        """Load the full message history for a thread as ModelMessage[]."""
        await self._init_db()
        try:
            async with aiosqlite.connect(self.db_path) as db:
                async with db.execute(
                    "SELECT messages_json FROM threads WHERE id = ?",
                    (thread_id,),
                ) as cur:
                    row = await cur.fetchone()
                    if not row:
                        return []
                    messages_blob: bytes = row[0]
                    return ModelMessagesTypeAdapter.validate_json(messages_blob)
        except Exception as e:  # pragma: no cover
            logger.warning(
                "threads.get_messages_failed", thread_id=thread_id, error=str(e)
            )
            return []

    async def list_threads(
        self, *, database_name: str | None = None, limit: int = 50
    ) -> list[Thread]:
        await self._init_db()
        try:
            query = (
                "SELECT id, database_name, title, created_at, ended_at, last_activity_at, model_name"
                " FROM threads"
            )
            params: list[Any] = []
            if database_name:
                query += " WHERE database_name = ?"
                params.append(database_name)
            query += " ORDER BY last_activity_at DESC LIMIT ?"
            params.append(limit)

            async with aiosqlite.connect(self.db_path) as db:
                async with db.execute(query, params) as cur:
                    threads: list[Thread] = []
                    async for row in cur:
                        threads.append(
                            Thread(
                                id=row[0],
                                database_name=row[1],
                                title=row[2],
                                created_at=row[3],
                                ended_at=row[4],
                                last_activity_at=row[5],
                                model_name=row[6],
                            )
                        )
                    return threads
        except Exception as e:  # pragma: no cover
            logger.warning("threads.list_failed", error=str(e))
            return []

    async def delete_thread(self, thread_id: str) -> bool:
        await self._init_db()
        try:
            async with self._lock, aiosqlite.connect(self.db_path) as db:
                cur = await db.execute("DELETE FROM threads WHERE id = ?", (thread_id,))
                await db.commit()
                deleted = cur.rowcount > 0
                if deleted:
                    logger.info("threads.delete", thread_id=thread_id)
                return deleted
        except Exception as e:  # pragma: no cover
            logger.warning("threads.delete_failed", thread_id=thread_id, error=str(e))
            return False

    async def prune_threads(self, older_than_days: int = 30) -> int:
        """Delete threads whose last_activity_at is older than the cutoff.

        Args:
            older_than_days: Threads with last_activity_at older than this many days ago will be deleted.

        Returns:
            The number of rows deleted (best-effort; 0 on failure).
        """
        await self._init_db()
        cutoff = time.time() - older_than_days * 24 * 3600
        try:
            async with self._lock, aiosqlite.connect(self.db_path) as db:
                cur = await db.execute(
                    "DELETE FROM threads WHERE last_activity_at < ?",
                    (cutoff,),
                )
                await db.commit()
                deleted = cur.rowcount or 0
                logger.info("threads.prune", days=older_than_days, deleted=deleted)
                return deleted
        except Exception as e:  # pragma: no cover
            logger.warning("threads.prune_failed", days=older_than_days, error=str(e))
            return 0



"""Authentication configuration management for SQLSaber.

This module currently tracks whether the user has run the interactive auth setup.
SQLSaber authenticates to providers via API keys (environment variables or OS
keyring storage).
"""

import json
import os
import platform
import stat
from enum import Enum
from pathlib import Path
from typing import Any

import platformdirs


class AuthMethod(Enum):
    """Authentication methods available in SQLSaber."""

    API_KEY = "api_key"


class AuthConfigManager:
    """Manages authentication configuration persistence."""

    def __init__(self):
        self.config_dir = Path(platformdirs.user_config_dir("sqlsaber", "sqlsaber"))
        self.config_file = self.config_dir / "auth_config.json"
        self._ensure_config_dir()

    def _ensure_config_dir(self) -> None:
        """Ensure config directory exists with proper permissions."""
        self.config_dir.mkdir(parents=True, exist_ok=True)
        self._set_secure_permissions(self.config_dir, is_directory=True)

    def _set_secure_permissions(self, path: Path, is_directory: bool = False) -> None:
        """Set secure permissions cross-platform."""
        try:
            if platform.system() == "Windows":
                return

            if is_directory:
                os.chmod(path, stat.S_IRWXU)  # 0o700
            else:
                os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)  # 0o600
        except (OSError, PermissionError):
            pass

    def _load_config(self) -> dict[str, Any]:
        """Load configuration from file."""
        if not self.config_file.exists():
            return {"auth_method": None}

        try:
            with open(self.config_file, "r") as f:
                return json.load(f)
        except (json.JSONDecodeError, IOError):
            return {"auth_method": None}

    def _save_config(self, config: dict[str, Any]) -> None:
        """Save configuration to file."""
        with open(self.config_file, "w") as f:
            json.dump(config, f, indent=2)

        self._set_secure_permissions(self.config_file, is_directory=False)

    def get_auth_method(self) -> AuthMethod | None:
        """Get the configured authentication method."""
        config = self._load_config()
        auth_method_str = config.get("auth_method")

        if auth_method_str is None:
            return None

        try:
            return AuthMethod(auth_method_str)
        except ValueError:
            return None

    def set_auth_method(self, auth_method: AuthMethod) -> None:
        """Set the authentication method."""
        config = self._load_config()
        config["auth_method"] = auth_method.value
        self._save_config(config)

    def clear_auth_method(self) -> None:
        """Clear any configured authentication method."""
        config = self._load_config()
        config["auth_method"] = None
        self._save_config(config)

    def has_auth_configured(self) -> bool:
        """Check if authentication method is configured."""
        return self.get_auth_method() is not None



from __future__ import annotations

import pytest

from sqlsaber import SQLSaber
from sqlsaber.memory.manager import MemoryManager


@pytest.mark.asyncio
async def test_api_memory_text_injected(temp_dir, monkeypatch):
    config_dir = temp_dir / "config"
    monkeypatch.setattr(
        "platformdirs.user_config_dir", lambda *args, **kwargs: str(config_dir)
    )

    saber = SQLSaber(
        database="sqlite:///:memory:",
        model_name="anthropic:claude-3-5-sonnet",
        api_key="test-key",
        memory="hello memory",
    )

    try:
        prompt = saber.agent.system_prompt_text(include_memory=True)
        assert "hello memory" in prompt
    finally:
        await saber.close()


@pytest.mark.asyncio
async def test_api_memory_file_injected(temp_dir, monkeypatch):
    config_dir = temp_dir / "config"
    monkeypatch.setattr(
        "platformdirs.user_config_dir", lambda *args, **kwargs: str(config_dir)
    )

    memory_file = temp_dir / "memory.txt"
    memory_file.write_text("memory from file", encoding="utf-8")

    saber = SQLSaber(
        database="sqlite:///:memory:",
        model_name="anthropic:claude-3-5-sonnet",
        api_key="test-key",
        memory=memory_file,
    )

    try:
        prompt = saber.agent.system_prompt_text(include_memory=True)
        assert "memory from file" in prompt
    finally:
        await saber.close()


@pytest.mark.asyncio
async def test_api_empty_memory_disables_saved_memories(temp_dir, monkeypatch):
    config_dir = temp_dir / "config"
    monkeypatch.setattr(
        "platformdirs.user_config_dir", lambda *args, **kwargs: str(config_dir)
    )

    db_path = temp_dir / "my.db"
    db_path.write_text("", encoding="utf-8")

    memory_manager = MemoryManager()
    memory_manager.add_memory("my", "saved-memory")

    saber = SQLSaber(
        database=f"sqlite:///{db_path}",
        model_name="anthropic:claude-3-5-sonnet",
        api_key="test-key",
        memory="",
    )

    try:
        prompt = saber.agent.system_prompt_text(include_memory=True)
        assert "saved-memory" not in prompt
    finally:
        await saber.close()



"""Tests for the interactive auth setup flow."""

from __future__ import annotations

from collections import deque
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from sqlsaber.application.auth_setup import setup_auth


class DummyPrompter:
    """Simple prompter stub that yields predefined responses."""

    def __init__(
        self,
        selects: list[Any] | None = None,
        confirms: list[bool | None] | None = None,
    ):
        self._selects = deque(selects or [])
        self._confirms = deque(confirms or [])

    async def select(
        self,
        message: str,
        choices: list[Any] | None = None,
        default: Any = None,
        use_search_filter: bool = False,
        use_jk_keys: bool = True,
    ) -> Any:
        try:
            return self._selects.popleft()
        except IndexError:
            return None

    async def confirm(self, message: str, default: bool = False) -> bool | None:
        try:
            return self._confirms.popleft()
        except IndexError:
            return None

    async def text(self, message: str, default: str = "", validate=None) -> str | None:
        return None

    async def path(self, message: str, only_directories: bool = False) -> str | None:
        return None


@pytest.mark.asyncio
async def test_setup_auth_resets_existing_api_key(monkeypatch: pytest.MonkeyPatch):
    """Stored API key is cleared when the user opts into reset."""
    prompter = DummyPrompter(selects=["openai"], confirms=[True])
    auth_manager = MagicMock()
    api_key_manager = MagicMock()
    api_key_manager.get_env_var_name.return_value = "OPENAI_API_KEY"
    api_key_manager.has_stored_api_key.return_value = True
    api_key_manager.delete_api_key.return_value = True

    configure_api_key = AsyncMock(return_value=True)

    with patch("sqlsaber.application.auth_setup.configure_api_key", configure_api_key):
        success, provider = await setup_auth(
            prompter=prompter,
            auth_manager=auth_manager,
            api_key_manager=api_key_manager,
            default_provider="openai",
        )

    assert success is True
    assert provider == "openai"
    api_key_manager.delete_api_key.assert_called_once_with("openai")
    configure_api_key.assert_awaited_once()



"""Multiple-CSV database connection using DuckDB backend.

This connection creates one DuckDB view per CSV file for each query.
"""

import asyncio
from collections import Counter
from typing import Any
from urllib.parse import parse_qs, urlparse

import duckdb

from .base import DEFAULT_QUERY_TIMEOUT, BaseDatabaseConnection, QueryTimeoutError
from .csv import CSVConnection, _execute_duckdb_transaction


class CSVsConnection(BaseDatabaseConnection):
    """Connection that exposes multiple CSV files as DuckDB views."""

    def __init__(self, connection_string: str):
        super().__init__(connection_string)

        parsed = urlparse(connection_string)
        if parsed.scheme != "csvs":
            raise ValueError(
                f"Invalid connection string for CSVsConnection: {connection_string}"
            )

        params = parse_qs(parsed.query)
        specs = [spec for spec in params.get("spec", []) if spec.strip()]
        if not specs:
            raise ValueError(
                "csvs connection string must include one or more 'spec=' parameters"
            )

        sources = [CSVConnection(spec) for spec in specs]
        self._dedupe_view_names(sources)

        # Used by DuckDBSchemaIntrospector to route through execute_query.
        self.csv_sources: list[CSVConnection] = sources

    @staticmethod
    def _dedupe_view_names(sources: list[CSVConnection]) -> None:
        counts = Counter(src.table_name for src in sources)
        used: dict[str, int] = {}

        for src in sources:
            base = src.table_name
            if counts[base] <= 1:
                continue

            used[base] = used.get(base, 0) + 1
            if used[base] == 1:
                continue
            src.table_name = f"{base}_{used[base]}"

    @property
    def sqlglot_dialect(self) -> str:
        return "duckdb"

    @property
    def display_name(self) -> str:
        return "DuckDB"

    async def get_pool(self):
        return None

    async def close(self):
        pass

    async def execute_query(
        self, query: str, *args, timeout: float | None = None, commit: bool = False
    ) -> list[dict[str, Any]]:
        effective_timeout = timeout or DEFAULT_QUERY_TIMEOUT
        args_tuple = tuple(args) if args else tuple()

        def _run_query() -> list[dict[str, Any]]:
            conn = duckdb.connect(":memory:")
            try:
                for src in self.csv_sources:
                    src._create_view(conn)
                return _execute_duckdb_transaction(conn, query, args_tuple, commit)
            finally:
                conn.close()

        try:
            return await asyncio.wait_for(
                asyncio.to_thread(_run_query), timeout=effective_timeout
            )
        except asyncio.TimeoutError as exc:
            raise QueryTimeoutError(effective_timeout or 0) from exc



"""HTML export utilities for thread transcripts."""

import json
from html import escape

from pydantic_ai.messages import ModelMessage

from sqlsaber.threads.storage import Thread


def build_turn_slices(all_msgs: list[ModelMessage]) -> list[tuple[int, int]]:
    """Return (start, end) slices for turns, grouped by user-prompt messages."""
    user_indices: list[int] = []
    for idx, message in enumerate(all_msgs):
        for part in getattr(message, "parts", []):
            if getattr(part, "part_kind", "") == "user-prompt":
                user_indices.append(idx)
                break

    slices: list[tuple[int, int]] = []
    if user_indices:
        for i, start_idx in enumerate(user_indices):
            end_idx = (
                user_indices[i + 1] if i + 1 < len(user_indices) else len(all_msgs)
            )
            slices.append((start_idx, end_idx))
    else:
        slices = [(0, len(all_msgs))]
    return slices


def _html_escape_multimodal(content: object) -> str:
    """Flatten and HTML-escape multimodal user content (strings + JSON dumps)."""
    if isinstance(content, str):
        return escape(content)
    if isinstance(content, list):
        parts: list[str] = []
        for seg in content:
            if isinstance(seg, str):
                parts.append(seg)
            else:
                try:
                    parts.append(json.dumps(seg, ensure_ascii=False))
                except Exception:
                    parts.append(str(seg))
        text = "\n".join(s for s in parts if s)
        return escape(text)
    return escape(str(content))


def _render_sql_results_table(content_str: str, sql_query: str | None = None) -> str:
    """Render SQL execute results as an HTML table with the query."""
    parts: list[str] = []

    if sql_query:
        parts.append(
            f'<div class="sql-query"><pre><code class="language-sql">{escape(sql_query)}</code></pre></div>'
        )

    try:
        data = json.loads(content_str)
        if isinstance(data, dict) and data.get("success") and data.get("results"):
            results = data["results"]
            if isinstance(results, list) and results:
                first = results[0]
                if isinstance(first, dict):
                    headers = list(first.keys())
                    rows_html = []
                    for row in results[:100]:  # Limit to 100 rows
                        cells = "".join(
                            f"<td>{escape(str(row.get(h, '')))}</td>" for h in headers
                        )
                        rows_html.append(f"<tr>{cells}</tr>")
                    header_html = "".join(f"<th>{escape(h)}</th>" for h in headers)
                    count_note = (
                        f'<p class="result-count">{len(results)} row(s) returned</p>'
                    )
                    if len(results) > 100:
                        count_note = f'<p class="result-count">Showing 100 of {len(results)} rows</p>'
                    parts.append(f"""
{count_note}
<div class="table-wrapper">
<table class="sql-results">
<thead><tr>{header_html}</tr></thead>
<tbody>{"".join(rows_html)}</tbody>
</table>
</div>""")
                    return "".join(parts)
            elif isinstance(results, list) and not results:
                parts.append('<p class="result-count">0 rows returned</p>')
                return "".join(parts)
        if isinstance(data, dict) and "error" in data:
            error = escape(str(data.get("error", "")))
            parts.append(
                f'<div class="sql-error"><strong>Error:</strong> {error}</div>'
            )
            return "".join(parts)
    except (json.JSONDecodeError, TypeError):
        pass

    parts.append(f"<pre><code>{escape(content_str)}</code></pre>")
    return "".join(parts)


def _render_list_tables_html(content_str: str) -> str:
    """Render list_tables results as an HTML table."""
    try:
        data = json.loads(content_str)

        if isinstance(data, dict) and "error" in data:
            return f'<div class="sql-error"><strong>Error:</strong> {escape(str(data["error"]))}</div>'

        if isinstance(data, list):
            if not data:
                return '<p class="result-count">No tables found in the database.</p>'
            if isinstance(data[0], str):
                rows_html = [
                    f"<tr><td>—</td><td>{escape(t)}</td><td>—</td></tr>" for t in data
                ]
                return f"""
<p class="result-count">{len(data)} table(s) found</p>
<div class="table-wrapper">
<table class="sql-results">
<thead><tr><th>Schema</th><th>Table Name</th><th>Type</th></tr></thead>
<tbody>{"".join(rows_html)}</tbody>
</table>
</div>"""

        if isinstance(data, dict):
            tables = data.get("tables", [])
            total = data.get("total_tables", len(tables))

            if not tables:
                return '<p class="result-count">No tables found in the database.</p>'

            rows_html = []
            for t in tables:
                schema = escape(str(t.get("schema", "")))
                name = escape(str(t.get("name", "")))
                ttype = escape(str(t.get("type", "")))
                rows_html.append(
                    f"<tr><td>{schema}</td><td>{name}</td><td>{ttype}</td></tr>"
                )

            return f"""
<p class="result-count">{total} table(s) found</p>
<div class="table-wrapper">
<table class="sql-results">
<thead><tr><th>Schema</th><th>Table Name</th><th>Type</th></tr></thead>
<tbody>{"".join(rows_html)}</tbody>
</table>
</div>"""

        return f"<pre><code>{escape(content_str)}</code></pre>"
    except (json.JSONDecodeError, TypeError, AttributeError):
        return f"<pre><code>{escape(content_str)}</code></pre>"


def _render_introspect_schema_html(content_str: str) -> str:
    """Render introspect_schema results as formatted HTML tables."""
    try:
        data = json.loads(content_str)
        if "error" in data:
            return f'<div class="sql-error"><strong>Error:</strong> {escape(str(data["error"]))}</div>'

        if not data:
            return '<p class="result-count">No schema information found.</p>'

        parts: list[str] = []
        parts.append(f'<p class="result-count">{len(data)} table(s) introspected</p>')

        for table_name, table_info in data.items():
            parts.append('<div class="schema-table">')
            parts.append(f'<h4 class="table-name">{escape(table_name)}</h4>')

            if table_info.get("comment"):
                parts.append(
                    f'<p class="table-comment">{escape(str(table_info["comment"]))}</p>'
                )

            columns = table_info.get("columns", {})
            if columns:
                rows = []
                for col_name, col_info in columns.items():
                    nullable = "✓" if col_info.get("nullable") else "✗"
                    default = escape(str(col_info.get("default") or "—"))
                    comment = escape(str(col_info.get("comment") or ""))
                    rows.append(
                        f"<tr><td>{escape(col_name)}</td><td>{escape(str(col_info.get('type', '')))}</td>"
                        f"<td>{nullable}</td><td>{default}</td><td>{comment}</td></tr>"
                    )
                parts.append("""
<div class="table-wrapper">
<table class="sql-results schema-columns">
<thead><tr><th>Column</th><th>Type</th><th>Nullable</th><th>Default</th><th>Comment</th></tr></thead>
<tbody>""")
                parts.append("".join(rows))
                parts.append("</tbody></table></div>")

            pks = table_info.get("primary_keys", [])
            if pks:
                parts.append(
                    f'<p class="key-info"><strong>Primary Keys:</strong> {escape(", ".join(pks))}</p>'
                )

            fks = table_info.get("foreign_keys", [])
            if fks:
                fk_list = "".join(f"<li>{escape(fk)}</li>" for fk in fks)
                parts.append(
                    f'<p class="key-info"><strong>Foreign Keys:</strong></p><ul class="key-list">{fk_list}</ul>'
                )

            parts.append("</div>")

        return "".join(parts)
    except (json.JSONDecodeError, TypeError):
        return f"<pre><code>{escape(content_str)}</code></pre>"


def _human_readable_time(timestamp: float | None) -> str:
    """Format a timestamp as a human-readable string."""
    import time

    if not timestamp:
        return "-"
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp))


def render_thread_html(thread: Thread, all_msgs: list[ModelMessage]) -> str:
    """Generate a standalone HTML document for a thread transcript."""
    title = thread.title or f"SQLsaber thread {thread.id}"
    escaped_title = escape(title)

    head = f"""<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>{escaped_title}</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="preconnect" href="https://fonts.googleapis.com">
  <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
  <link href="https://fonts.googleapis.com/css2?family=IBM+Plex+Mono:wght@400;500;600&family=IBM+Plex+Sans:wght@400;500;600&display=swap" rel="stylesheet">
  <link id="hljs-theme" rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/github-dark.min.css">
  <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/highlight.min.js"></script>
  <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/languages/sql.min.js"></script>
  <script src="https://cdn.jsdelivr.net/npm/marked/marked.min.js"></script>
  <script>
    (function() {{
      const saved = localStorage.getItem('sqlsaber-theme');
      if (saved === 'light') document.documentElement.setAttribute('data-theme', 'light');
    }})();
  </script>
  <style>
    :root {{
      --bg-primary: #0d1117;
      --bg-secondary: #161b22;
      --bg-tertiary: #21262d;
      --border-color: #30363d;
      --text-primary: #e6edf3;
      --text-secondary: #8b949e;
      --text-muted: #6e7681;
      --accent-blue: #58a6ff;
      --accent-green: #3fb950;
      --accent-amber: #d29922;
      --accent-red: #f85149;
      --accent-purple: #a371f7;
      --font-sans: 'IBM Plex Sans', -apple-system, BlinkMacSystemFont, sans-serif;
      --font-mono: 'IBM Plex Mono', 'SF Mono', Consolas, monospace;
    }}
    [data-theme="light"] {{
      --bg-primary: #ffffff;
      --bg-secondary: #f6f8fa;
      --bg-tertiary: #eaeef2;
      --border-color: #d0d7de;
      --text-primary: #1f2328;
      --text-secondary: #59636e;
      --text-muted: #8c959f;
      --accent-blue: #0969da;
      --accent-green: #1a7f37;
      --accent-amber: #9a6700;
      --accent-red: #d1242f;
      --accent-purple: #8250df;
    }}
    * {{ box-sizing: border-box; }}
    html {{ font-size: 15px; }}
    body {{
      font-family: var(--font-sans);
      background: var(--bg-primary);
      color: var(--text-primary);
      line-height: 1.6;
      margin: 0;
      padding: 2rem 1rem;
    }}
    .container {{
      max-width: 900px;
      margin: 0 auto;
    }}
    header {{
      border-bottom: 1px solid var(--border-color);
      padding-bottom: 1.5rem;
      margin-bottom: 2rem;
    }}
    .logo {{
      display: flex;
      align-items: center;
      gap: 0.5rem;
      margin-bottom: 1rem;
    }}
    .logo svg {{
      width: 28px;
      height: 28px;
    }}
    .logo span {{
      font-family: var(--font-mono);
      font-weight: 600;
      font-size: 0.85rem;
      color: var(--text-secondary);
      letter-spacing: 0.02em;
    }}
    h1 {{
      font-size: 1.75rem;
      font-weight: 600;
      margin: 0 0 0.75rem 0;
      color: var(--text-primary);
    }}
    .thread-id {{
      font-family: var(--font-mono);
      font-size: 0.75rem;
      color: var(--text-muted);
      background: var(--bg-tertiary);
      padding: 0.25rem 0.5rem;
      border-radius: 4px;
      display: inline-block;
      margin-bottom: 1rem;
    }}
    .meta-grid {{
      display: grid;
      grid-template-columns: repeat(auto-fit, minmax(180px, 1fr));
      gap: 0.75rem;
    }}
    .meta-item {{
      background: var(--bg-secondary);
      border: 1px solid var(--border-color);
      border-radius: 6px;
      padding: 0.75rem 1rem;
    }}
    .meta-label {{
      font-size: 0.7rem;
      text-transform: uppercase;
      letter-spacing: 0.05em;
      color: var(--text-muted);
      margin-bottom: 0.25rem;
    }}
    .meta-value {{
      font-size: 0.9rem;
      color: var(--text-primary);
      font-weight: 500;
    }}
    .turn {{
      margin-bottom: 1.5rem;
      border-radius: 8px;
      overflow: hidden;
      border: 1px solid var(--border-color);
    }}
    .turn-header {{
      display: flex;
      align-items: center;
      gap: 0.5rem;
      padding: 0.75rem 1rem;
      font-size: 0.8rem;
      font-weight: 600;
      text-transform: uppercase;
      letter-spacing: 0.03em;
    }}
    .turn.user .turn-header {{
      background: linear-gradient(135deg, rgba(88,166,255,0.15), rgba(88,166,255,0.05));
      color: var(--accent-blue);
      border-bottom: 1px solid rgba(88,166,255,0.2);
    }}
    .turn.assistant .turn-header {{
      background: linear-gradient(135deg, rgba(63,185,80,0.15), rgba(63,185,80,0.05));
      color: var(--accent-green);
      border-bottom: 1px solid rgba(63,185,80,0.2);
    }}
    .turn-content {{
      padding: 1rem 1.25rem;
      background: var(--bg-secondary);
    }}
    .turn.user .turn-content {{
      font-size: 1rem;
      white-space: pre-wrap;
    }}
    .markdown-content {{
      font-size: 0.95rem;
    }}
    .markdown-content h1, .markdown-content h2, .markdown-content h3 {{
      color: var(--text-primary);
      margin-top: 1.5rem;
      margin-bottom: 0.75rem;
    }}
    .markdown-content h1 {{ font-size: 1.4rem; }}
    .markdown-content h2 {{ font-size: 1.2rem; }}
    .markdown-content h3 {{ font-size: 1.05rem; }}
    .markdown-content p {{ margin: 0.75rem 0; }}
    .markdown-content ul, .markdown-content ol {{
      margin: 0.75rem 0;
      padding-left: 1.5rem;
    }}
    .markdown-content li {{ margin: 0.25rem 0; }}
    .markdown-content code {{
      font-family: var(--font-mono);
      background: var(--bg-tertiary);
      padding: 0.15rem 0.4rem;
      border-radius: 4px;
      font-size: 0.85em;
    }}
    .markdown-content pre {{
      background: var(--bg-primary);
      border: 1px solid var(--border-color);
      border-radius: 6px;
      padding: 1rem;
      overflow-x: auto;
      margin: 1rem 0;
    }}
    .markdown-content pre code {{
      background: none;
      padding: 0;
      font-size: 0.85rem;
      line-height: 1.5;
    }}
    .markdown-content a {{
      color: var(--accent-blue);
      text-decoration: none;
    }}
    .markdown-content a:hover {{ text-decoration: underline; }}
    .markdown-content blockquote {{
      border-left: 3px solid var(--accent-purple);
      margin: 1rem 0;
      padding-left: 1rem;
      color: var(--text-secondary);
    }}
    .tool-section {{
      margin-top: 1rem;
      border: 1px solid var(--border-color);
      border-radius: 6px;
      overflow: hidden;
    }}
    .tool-header {{
      display: flex;
      align-items: center;
      gap: 0.5rem;
      padding: 0.6rem 1rem;
      background: var(--bg-tertiary);
      cursor: pointer;
      font-size: 0.8rem;
      color: var(--text-secondary);
      border: none;
      width: 100%;
      text-align: left;
    }}
    .tool-header:hover {{ background: var(--bg-primary); }}
    .tool-icon {{
      color: var(--accent-amber);
    }}
    .tool-name {{
      font-family: var(--font-mono);
      color: var(--accent-amber);
      font-weight: 500;
    }}
    .tool-body {{
      padding: 1rem;
      background: var(--bg-secondary);
      border-top: 1px solid var(--border-color);
    }}
    details.tool > summary {{
      list-style: none;
    }}
    details.tool > summary::-webkit-details-marker {{
      display: none;
    }}
    details.tool > summary::before {{
      content: '▸';
      margin-right: 0.5rem;
      transition: transform 0.2s;
      display: inline-block;
    }}
    details.tool[open] > summary::before {{
      transform: rotate(90deg);
    }}
    .table-wrapper {{
      overflow-x: auto;
      margin: 0.5rem 0;
    }}
    .sql-results {{
      width: 100%;
      border-collapse: collapse;
      font-family: var(--font-mono);
      font-size: 0.8rem;
    }}
    .sql-results th {{
      background: var(--bg-tertiary);
      color: var(--accent-blue);
      font-weight: 600;
      text-align: left;
      padding: 0.6rem 0.75rem;
      border-bottom: 2px solid var(--accent-blue);
      white-space: nowrap;
    }}
    .sql-results td {{
      padding: 0.5rem 0.75rem;
      border-bottom: 1px solid var(--border-color);
      color: var(--text-primary);
    }}
    .sql-results tr:hover td {{
      background: var(--bg-tertiary);
    }}
    .sql-results tr:last-child td {{
      border-bottom: none;
    }}
    .result-count {{
      font-size: 0.75rem;
      color: var(--text-muted);
      margin: 0 0 0.5rem 0;
      font-family: var(--font-mono);
    }}
    .sql-error {{
      background: rgba(248, 81, 73, 0.1);
      border: 1px solid rgba(248, 81, 73, 0.3);
      border-radius: 6px;
      padding: 0.75rem 1rem;
      color: var(--accent-red);
      font-size: 0.9rem;
    }}
    .sql-query {{
      margin-bottom: 0.75rem;
    }}
    .sql-query pre {{
      margin: 0;
      background: var(--bg-primary);
      border: 1px solid var(--border-color);
      border-radius: 6px;
      padding: 0.75rem 1rem;
    }}
    .sql-query code {{
      font-family: var(--font-mono);
      font-size: 0.85rem;
      line-height: 1.5;
    }}
    .schema-table {{
      margin-bottom: 1.5rem;
      padding-bottom: 1rem;
      border-bottom: 1px solid var(--border-color);
    }}
    .schema-table:last-child {{
      border-bottom: none;
      margin-bottom: 0;
    }}
    .table-name {{
      color: var(--accent-purple);
      font-family: var(--font-mono);
      font-size: 1rem;
      margin: 0 0 0.5rem 0;
    }}
    .table-comment {{
      color: var(--text-secondary);
      font-size: 0.85rem;
      margin: 0 0 0.75rem 0;
      font-style: italic;
    }}
    .key-info {{
      font-size: 0.8rem;
      color: var(--text-secondary);
      margin: 0.5rem 0 0 0;
    }}
    .key-info strong {{
      color: var(--accent-amber);
    }}
    .key-list {{
      margin: 0.25rem 0 0 1.25rem;
      padding: 0;
      font-size: 0.8rem;
      color: var(--text-secondary);
    }}
    .key-list li {{
      margin: 0.15rem 0;
    }}
    .empty-state {{
      text-align: center;
      padding: 3rem;
      color: var(--text-muted);
    }}
    footer {{
      margin-top: 3rem;
      padding-top: 1.5rem;
      border-top: 1px solid var(--border-color);
      text-align: center;
      font-size: 0.75rem;
      color: var(--text-muted);
    }}
    footer a {{ color: var(--accent-blue); text-decoration: none; }}
    footer a:hover {{ text-decoration: underline; }}
    .header-row {{
      display: flex;
      justify-content: space-between;
      align-items: center;
      margin-bottom: 1rem;
    }}
    .theme-toggle {{
      background: var(--bg-tertiary);
      border: 1px solid var(--border-color);
      border-radius: 6px;
      padding: 0.5rem;
      cursor: pointer;
      color: var(--text-secondary);
      display: flex;
      align-items: center;
      justify-content: center;
      transition: background 0.2s, color 0.2s;
    }}
    .theme-toggle:hover {{
      background: var(--bg-secondary);
      color: var(--text-primary);
    }}
    .theme-toggle svg {{
      width: 18px;
      height: 18px;
    }}
    .icon-sun {{ display: none; }}
    .icon-moon {{ display: block; }}
    [data-theme="light"] .icon-sun {{ display: block; }}
    [data-theme="light"] .icon-moon {{ display: none; }}
    @media (max-width: 600px) {{
      html {{ font-size: 14px; }}
      body {{ padding: 1rem 0.75rem; }}
      .meta-grid {{ grid-template-columns: 1fr 1fr; }}
    }}
  </style>
</head>
"""

    meta_html = f"""
<body>
<div class="container">
  <header>
    <div class="header-row">
      <div class="logo">
        <svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
          <path d="M12 2L2 7l10 5 10-5-10-5z"/>
          <path d="M2 17l10 5 10-5"/>
          <path d="M2 12l10 5 10-5"/>
        </svg>
        <span>SQLsaber</span>
      </div>
      <button class="theme-toggle" id="theme-toggle" aria-label="Toggle theme">
        <svg class="icon-moon" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
          <path d="M21 12.79A9 9 0 1 1 11.21 3 7 7 0 0 0 21 12.79z"/>
        </svg>
        <svg class="icon-sun" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
          <circle cx="12" cy="12" r="5"/>
          <line x1="12" y1="1" x2="12" y2="3"/>
          <line x1="12" y1="21" x2="12" y2="23"/>
          <line x1="4.22" y1="4.22" x2="5.64" y2="5.64"/>
          <line x1="18.36" y1="18.36" x2="19.78" y2="19.78"/>
          <line x1="1" y1="12" x2="3" y2="12"/>
          <line x1="21" y1="12" x2="23" y2="12"/>
          <line x1="4.22" y1="19.78" x2="5.64" y2="18.36"/>
          <line x1="18.36" y1="5.64" x2="19.78" y2="4.22"/>
        </svg>
      </button>
    </div>
    <h1>{escaped_title}</h1>
    <code class="thread-id">{escape(thread.id)}</code>
    <div class="meta-grid">
      <div class="meta-item">
        <div class="meta-label">Database</div>
        <div class="meta-value">{escape(thread.database_name or "—")}</div>
      </div>
      <div class="meta-item">
        <div class="meta-label">Model</div>
        <div class="meta-value">{escape(thread.model_name or "—")}</div>
      </div>
      <div class="meta-item">
        <div class="meta-label">Created</div>
        <div class="meta-value">{escape(_human_readable_time(thread.created_at))}</div>
      </div>
      <div class="meta-item">
        <div class="meta-label">Last Activity</div>
        <div class="meta-value">{escape(_human_readable_time(thread.last_activity_at))}</div>
      </div>
    </div>
  </header>
  <main>
"""

    body_parts: list[str] = []
    slices = build_turn_slices(all_msgs)

    for start_idx, end_idx in slices:
        if not (0 <= start_idx < len(all_msgs)):
            continue

        user_msg = all_msgs[start_idx]
        user_text = ""
        for part in getattr(user_msg, "parts", []):
            if getattr(part, "part_kind", "") == "user-prompt":
                content = getattr(part, "content", None)
                user_text = _html_escape_multimodal(content)
                break

        body_parts.append(
            f"""<section class="turn user">
  <div class="turn-header">
    <svg width="16" height="16" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
      <path d="M20 21v-2a4 4 0 0 0-4-4H8a4 4 0 0 0-4 4v2"/>
      <circle cx="12" cy="7" r="4"/>
    </svg>
    User
  </div>
  <div class="turn-content">{user_text or "(no content)"}</div>
</section>
"""
        )

        tool_calls: dict[str, dict[str, str]] = {}
        for i in range(start_idx + 1, end_idx):
            msg = all_msgs[i]
            for part in getattr(msg, "parts", []):
                kind = getattr(part, "part_kind", "")
                if kind in ("tool-call", "builtin-tool-call"):
                    call_id = getattr(part, "tool_call_id", None)
                    tool_name = str(getattr(part, "tool_name", ""))
                    args = getattr(part, "args", {})
                    if call_id and tool_name:
                        args_dict: dict = {}
                        if isinstance(args, dict):
                            args_dict = args
                        elif isinstance(args, str):
                            try:
                                parsed = json.loads(args)
                                if isinstance(parsed, dict):
                                    args_dict = parsed
                            except Exception:
                                pass
                        sql_query = None
                        if tool_name == "execute_sql":
                            sql_query = args_dict.get("sql") or args_dict.get("query")
                        tool_calls[call_id] = {
                            "name": tool_name,
                            "sql": sql_query or "",
                        }

        for i in range(start_idx + 1, end_idx):
            msg = all_msgs[i]
            assistant_text_blocks: list[str] = []
            tool_details_html: list[str] = []

            for part in getattr(msg, "parts", []):
                kind = getattr(part, "part_kind", "")

                if kind == "text":
                    text = getattr(part, "content", "")
                    if isinstance(text, str) and text.strip():
                        assistant_text_blocks.append(escape(text))

                elif kind in ("tool-return", "builtin-tool-return"):
                    tool_name = str(getattr(part, "tool_name", "tool"))
                    call_id = getattr(part, "tool_call_id", None)
                    content = getattr(part, "content", None)
                    if isinstance(content, (dict, list)):
                        content_str = json.dumps(content, ensure_ascii=False, indent=2)
                    elif isinstance(content, str):
                        content_str = content
                    else:
                        content_str = json.dumps(
                            {"return_value": str(content)}, ensure_ascii=False, indent=2
                        )

                    sql_query = (
                        tool_calls.get(call_id, {}).get("sql") if call_id else None
                    )

                    if tool_name == "execute_sql":
                        result_html = _render_sql_results_table(content_str, sql_query)
                    elif tool_name == "list_tables":
                        result_html = _render_list_tables_html(content_str)
                    elif tool_name == "introspect_schema":
                        result_html = _render_introspect_schema_html(content_str)
                    else:
                        result_html = f"<pre><code>{escape(content_str)}</code></pre>"

                    tool_details_html.append(
                        f"""<details class="tool" open>
  <summary class="tool-header">
    <span class="tool-icon">⚡</span>
    <span class="tool-name">{escape(tool_name)}</span>
  </summary>
  <div class="tool-body">{result_html}</div>
</details>"""
                    )

            if assistant_text_blocks or tool_details_html:
                markdown_content = "\n\n".join(assistant_text_blocks)
                tool_html = "\n".join(tool_details_html)

                body_parts.append(
                    f"""<section class="turn assistant">
  <div class="turn-header">
    <svg width="16" height="16" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2">
      <path d="M12 2a10 10 0 1 0 10 10H12V2z"/>
      <path d="M12 2a7 7 0 0 1 7 7h-7V2z"/>
    </svg>
    Assistant
  </div>
  <div class="turn-content">
    <div class="markdown-content" data-markdown>{markdown_content}</div>
    {tool_html}
  </div>
</section>
"""
                )

    transcript_html = (
        "".join(body_parts)
        or '    <div class="empty-state"><p>No messages in this thread.</p></div>\n'
    )

    tail = """  </main>
  <footer>
    Exported from <a href="https://github.com/SarthakJariwala/sqlsaber">SQLsaber</a>
  </footer>
</div>
<script>
document.querySelectorAll('[data-markdown]').forEach(el => {
  const raw = el.textContent;
  el.innerHTML = marked.parse(raw);
});
hljs.highlightAll();

// Theme toggle
const toggle = document.getElementById('theme-toggle');
const hljsTheme = document.getElementById('hljs-theme');
function setTheme(theme) {
  if (theme === 'light') {
    document.documentElement.setAttribute('data-theme', 'light');
    hljsTheme.href = 'https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/github.min.css';
  } else {
    document.documentElement.removeAttribute('data-theme');
    hljsTheme.href = 'https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/github-dark.min.css';
  }
  localStorage.setItem('sqlsaber-theme', theme);
}
toggle.addEventListener('click', () => {
  const current = document.documentElement.getAttribute('data-theme');
  setTheme(current === 'light' ? 'dark' : 'light');
});
// Apply saved theme on load for hljs
if (localStorage.getItem('sqlsaber-theme') === 'light') {
  hljsTheme.href = 'https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/github.min.css';
}
</script>
</body>
</html>
"""

    return head + meta_html + transcript_html + tail



"""Tests for schema introspection."""

import aiosqlite
import duckdb
import pytest

from sqlsaber.database import DuckDBConnection, SQLiteConnection
from sqlsaber.database.schema import (
    DuckDBSchemaIntrospector,
    SchemaManager,
)


@pytest.mark.asyncio
async def test_duckdb_schema_manager(tmp_path):
    """Ensure DuckDB schema introspection surfaces tables and relationships."""
    db_path = tmp_path / "introspection.duckdb"

    conn = duckdb.connect(str(db_path))
    try:
        conn.execute("CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT);")
        conn.execute(
            "CREATE TABLE orders (id INTEGER, user_id INTEGER, FOREIGN KEY(user_id) REFERENCES users(id));"
        )
        conn.execute("CREATE UNIQUE INDEX idx_users_name ON users(name);")
    finally:
        conn.close()

    db_conn = DuckDBConnection(f"duckdb:///{db_path}")
    schema_manager = SchemaManager(db_conn)

    assert isinstance(schema_manager.introspector, DuckDBSchemaIntrospector)

    tables = await schema_manager.list_tables()
    table_names = {table["full_name"] for table in tables["tables"]}
    assert "main.users" in table_names
    assert "main.orders" in table_names

    schema_info = await schema_manager.get_schema_info()
    users_info = schema_info["main.users"]
    orders_info = schema_info["main.orders"]

    assert "id" in users_info["columns"]
    assert "INTEGER" in users_info["columns"]["id"]["data_type"].upper()
    assert "id" in users_info["primary_keys"]
    assert any(idx["name"] == "idx_users_name" for idx in users_info["indexes"])

    assert any(fk["column"] == "user_id" for fk in orders_info["foreign_keys"])


@pytest.mark.asyncio
async def test_duckdb_comments(tmp_path):
    """Test that DuckDB table and column comments are retrieved correctly."""
    db_path = tmp_path / "comments.duckdb"

    conn = duckdb.connect(str(db_path))
    try:
        conn.execute(
            "CREATE TABLE products (id INTEGER PRIMARY KEY, name TEXT, price DECIMAL);"
        )
        conn.execute("COMMENT ON TABLE products IS 'Product catalog';")
        conn.execute("COMMENT ON COLUMN products.id IS 'Unique product identifier';")
        conn.execute("COMMENT ON COLUMN products.name IS 'Product name';")
        conn.execute("COMMENT ON COLUMN products.price IS 'Product price in USD';")
    finally:
        conn.close()

    db_conn = DuckDBConnection(f"duckdb:///{db_path}")
    schema_manager = SchemaManager(db_conn)

    # Test list_tables includes comments
    tables = await schema_manager.list_tables()
    products_table = next(t for t in tables["tables"] if t["table_name"] == "products")
    assert products_table["table_comment"] == "Product catalog"

    # Test get_schema_info includes comments
    schema_info = await schema_manager.get_schema_info()
    products_info = schema_info["main.products"]

    assert products_info["comment"] == "Product catalog"
    assert products_info["columns"]["id"]["comment"] == "Unique product identifier"
    assert products_info["columns"]["name"]["comment"] == "Product name"
    assert products_info["columns"]["price"]["comment"] == "Product price in USD"


@pytest.mark.asyncio
async def test_sqlite_no_comments(tmp_path):
    """Verify that SQLite returns None for comments since it doesn't support them."""

    db_path = tmp_path / "no_comments.db"

    async with aiosqlite.connect(str(db_path)) as conn:
        await conn.execute("CREATE TABLE items (id INTEGER PRIMARY KEY, name TEXT);")
        await conn.commit()

    db_conn = SQLiteConnection(f"sqlite:///{db_path}")
    schema_manager = SchemaManager(db_conn)

    # Test list_tables has None for comments
    tables = await schema_manager.list_tables()
    items_table = next(t for t in tables["tables"] if t["table_name"] == "items")
    assert items_table["table_comment"] is None

    # Test get_schema_info has None for comments
    schema_info = await schema_manager.get_schema_info()
    items_info = schema_info["main.items"]

    assert items_info["comment"] is None
    assert items_info["columns"]["id"]["comment"] is None
    assert items_info["columns"]["name"]["comment"] is None



import tempfile
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from pydantic_ai.messages import (
    ModelMessagesTypeAdapter,
    ModelRequest,
    ModelResponse,
    TextPart,
    UserPromptPart,
)

from sqlsaber.threads.manager import ThreadManager
from sqlsaber.threads.storage import ThreadStorage


def _messages_bytes(user_text: str, assistant_text: str | None = None) -> bytes:
    parts = [UserPromptPart(user_text)]
    msgs = [ModelRequest(parts=parts)]
    if assistant_text:
        msgs.append(ModelResponse(parts=[TextPart(assistant_text)]))
    return ModelMessagesTypeAdapter.dump_json(msgs)


@pytest.fixture
def temp_storage():
    """Fixture for a real ThreadStorage backed by a temp file."""
    with tempfile.TemporaryDirectory() as tmp:
        store = ThreadStorage()
        store.db_path = Path(tmp) / "threads.db"
        yield store


@pytest.fixture
def thread_manager(temp_storage):
    """Fixture for ThreadManager using the temp storage."""
    return ThreadManager(storage=temp_storage)


@pytest.mark.asyncio
async def test_init_state(thread_manager):
    """Test initial state of ThreadManager."""
    assert thread_manager.current_thread_id is None
    assert thread_manager.first_message is True


@pytest.mark.asyncio
async def test_init_with_id(temp_storage):
    """Test initialization with an existing thread ID."""
    # Pre-create a thread in storage so it's "real"
    tm = ThreadManager(initial_thread_id="existing-id", storage=temp_storage)
    assert tm.current_thread_id == "existing-id"
    assert tm.first_message is False


@pytest.mark.asyncio
async def test_end_current_thread_active(thread_manager, temp_storage):
    """Test ending an active thread."""
    # Create a thread first
    messages_json = b"[]"
    thread_id = await temp_storage.save_snapshot(
        messages_json=messages_json, database_name="db1"
    )

    thread_manager.current_thread_id = thread_id

    result = await thread_manager.end_current_thread()

    assert result == thread_id

    # Verify in storage
    t = await temp_storage.get_thread(thread_id)
    assert t.ended_at is not None


@pytest.mark.asyncio
async def test_end_current_thread_none(thread_manager):
    """Test ending when no thread is active."""
    result = await thread_manager.end_current_thread()
    assert result is None


@pytest.mark.asyncio
async def test_clear_current_thread(thread_manager, temp_storage):
    """Test clearing the current thread."""
    # Create a thread first
    messages_json = b"[]"
    thread_id = await temp_storage.save_snapshot(
        messages_json=messages_json, database_name="db1"
    )

    thread_manager.current_thread_id = thread_id
    thread_manager.first_message = False

    await thread_manager.clear_current_thread()

    assert thread_manager.current_thread_id is None
    assert thread_manager.first_message is True

    # Check it was ended in storage
    t = await temp_storage.get_thread(thread_id)
    assert t.ended_at is not None


@pytest.mark.asyncio
async def test_save_run_new_thread(thread_manager, temp_storage):
    """Test saving a run for a new thread (first message)."""
    # Setup valid JSON for run_result
    json_bytes = _messages_bytes("hello")

    # Setup mocks for run_result
    run_result = MagicMock()
    run_result.all_messages_json.return_value = json_bytes
    run_result.all_messages.return_value = ["msg1"]

    # execute
    result = await thread_manager.save_run(
        run_result=run_result,
        database_name="db1",
        user_query="hello",
        model_name="gpt-4",
    )

    # Assertions
    assert result == ["msg1"]
    assert thread_manager.current_thread_id is not None
    assert thread_manager.first_message is False

    thread_id = thread_manager.current_thread_id

    # Verify in storage
    t = await temp_storage.get_thread(thread_id)
    assert t.database_name == "db1"
    assert t.title == "hello"
    assert t.model_name == "gpt-4"

    # Verify messages can be read back
    msgs = await temp_storage.get_thread_messages(thread_id)
    assert len(msgs) == 1


@pytest.mark.asyncio
async def test_save_run_existing_thread(thread_manager, temp_storage):
    """Test saving a run for an existing thread."""
    # Create initial thread
    thread_id = await temp_storage.save_snapshot(
        messages_json=_messages_bytes("original"), database_name="db1"
    )
    await temp_storage.save_metadata(
        thread_id=thread_id, title="original", model_name="gpt-4"
    )

    # Setup state
    thread_manager.current_thread_id = thread_id
    thread_manager.first_message = False

    # Setup valid JSON for updated run
    updated_bytes = _messages_bytes("original", "response")

    # Setup mocks
    run_result = MagicMock()
    run_result.all_messages_json.return_value = updated_bytes
    run_result.all_messages.return_value = ["msg1", "msg2"]

    # execute
    result = await thread_manager.save_run(
        run_result=run_result,
        database_name="db1",
        user_query="continuation",
        model_name="gpt-4",
    )

    # Assertions
    assert result == ["msg1", "msg2"]
    assert thread_manager.current_thread_id == thread_id

    # Verify messages updated in storage
    msgs = await temp_storage.get_thread_messages(thread_id)
    assert len(msgs) == 2

    # Verify title didn't change because first_message was False
    t = await temp_storage.get_thread(thread_id)
    assert t.title == "original"



"""SQL-related tools for database operations."""

import json

from sqlsaber.database import BaseDatabaseConnection
from sqlsaber.database.schema import SchemaManager

from .base import Tool
from .registry import register_tool
from .sql_guard import add_limit, validate_sql


class SQLTool(Tool):
    """Base class for SQL tools that need database access."""

    def __init__(
        self,
        db_connection: BaseDatabaseConnection | None = None,
        schema_manager: SchemaManager | None = None,
    ):
        """Initialize with optional database connection."""
        super().__init__()
        self.db = db_connection
        # allow_dangerous is set by SQLSaberAgent at session level
        # Do NOT expose this as a tool parameter to prevent LLM from escalating
        self.allow_dangerous: bool = False
        if schema_manager:
            self.schema_manager = schema_manager
        elif db_connection:
            self.schema_manager = SchemaManager(db_connection)
        else:
            self.schema_manager = None

    def set_connection(
        self,
        db_connection: BaseDatabaseConnection,
        schema_manager: SchemaManager | None = None,
    ) -> None:
        """Set the database connection after initialization."""
        self.db = db_connection
        if schema_manager:
            self.schema_manager = schema_manager
        else:
            self.schema_manager = SchemaManager(db_connection)


@register_tool
class ListTablesTool(SQLTool):
    """Tool for listing database tables."""

    @property
    def name(self) -> str:
        return "list_tables"

    async def execute(self) -> str:
        """List all tables in the database."""
        if not self.db or not self.schema_manager:
            return json.dumps({"error": "No database connection available"})

        try:
            tables_info = await self.schema_manager.list_tables()
            return json.dumps(tables_info)
        except Exception as e:
            return json.dumps({"error": f"Error listing tables: {str(e)}"})


@register_tool
class IntrospectSchemaTool(SQLTool):
    """Tool for introspecting database schema."""

    @property
    def name(self) -> str:
        return "introspect_schema"

    async def execute(self, table_pattern: str | None = None) -> str:
        """
        Introspect database schema.

        Args:
            table_pattern: Optional pattern to filter tables (e.g., 'public.users', 'user%', '%order%')
        """
        if not self.db or not self.schema_manager:
            return json.dumps({"error": "No database connection available"})

        try:
            schema_info = await self.schema_manager.get_schema_info(table_pattern)

            # Format the schema information
            formatted_info = {}
            for table_name, table_info in schema_info.items():
                table_data = {}

                # Add table comment if present
                if table_info.get("comment"):
                    table_data["comment"] = table_info["comment"]

                # Add columns with comments if present
                table_data["columns"] = {}
                for col_name, col_info in table_info["columns"].items():
                    column_data = {
                        "type": col_info["data_type"],
                        "nullable": col_info["nullable"],
                        "default": col_info["default"],
                    }
                    if col_info.get("comment"):
                        column_data["comment"] = col_info["comment"]
                    table_data["columns"][col_name] = column_data

                # Add other schema information
                table_data["primary_keys"] = table_info["primary_keys"]
                table_data["foreign_keys"] = [
                    f"{fk['column']} -> {fk['references']['table']}.{fk['references']['column']}"
                    for fk in table_info["foreign_keys"]
                ]
                table_data["indexes"] = [
                    f"{idx['name']} ({', '.join(idx['columns'])})"
                    + (" UNIQUE" if idx["unique"] else "")
                    + (f" [{idx['type']}]" if idx["type"] else "")
                    for idx in table_info["indexes"]
                ]

                formatted_info[table_name] = table_data

            return json.dumps(formatted_info)
        except Exception as e:
            return json.dumps({"error": f"Error introspecting schema: {str(e)}"})


@register_tool
class ExecuteSQLTool(SQLTool):
    """Tool for executing SQL queries."""

    DEFAULT_LIMIT = 100

    @property
    def name(self) -> str:
        return "execute_sql"

    async def execute(self, query: str, limit: int | None = 100) -> str:
        """
        Execute a SQL query against the database.

        Args:
            query: SQL query to execute
            limit: Maximum number of rows to return (default: 100)
        """
        if not self.db:
            return json.dumps({"error": "No database connection available"})

        if not query:
            return json.dumps({"error": "No query provided"})

        if limit is None:
            limit = self.DEFAULT_LIMIT

        try:
            # Get the dialect for this database
            dialect = self.db.sqlglot_dialect

            # Security check using sqlglot AST analysis
            validation_result = validate_sql(
                query, dialect, allow_dangerous=self.allow_dangerous
            )
            if not validation_result.allowed:
                return json.dumps({"error": validation_result.reason})

            # Add LIMIT if not present and it's a SELECT query
            if validation_result.is_select and limit:
                query = add_limit(query, dialect, limit)

            query_type = validation_result.query_type or "other"

            # Commit only for DML/DDL statements in dangerous mode
            commit = bool(self.allow_dangerous and query_type in {"dml", "ddl"})

            # Execute the query
            results = await self.db.execute_query(query, commit=commit)

            # Format response based on query type
            if query_type == "select":
                actual_limit = limit if limit is not None else len(results)
                return json.dumps(
                    {
                        "success": True,
                        "row_count": len(results),
                        "results": results[:actual_limit],
                        "truncated": len(results) > actual_limit,
                    }
                )
            elif query_type == "dml" or query_type == "ddl":
                return json.dumps({"success": True})
            else:
                return json.dumps(
                    {
                        "success": True,
                        "row_count": len(results),
                        "results": results,
                    }
                )

        except Exception as e:
            error_msg = str(e)

            # Provide helpful error messages
            suggestions = []
            if "column" in error_msg.lower() and "does not exist" in error_msg.lower():
                suggestions.append(
                    "Check column names using the schema introspection tool"
                )
            elif "table" in error_msg.lower() and "does not exist" in error_msg.lower():
                suggestions.append(
                    "Check table names using the schema introspection tool"
                )
            elif "syntax error" in error_msg.lower():
                suggestions.append(
                    "Review SQL syntax, especially JOIN conditions and WHERE clauses"
                )

            return json.dumps({"error": error_msg, "suggestions": suggestions})



"""Pydantic-AI Agent for SQLSaber.

This replaces the custom AnthropicSQLAgent and uses pydantic-ai's Agent,
function tools, and streaming event types directly.
"""

from collections.abc import AsyncIterable, Awaitable, Sequence
from typing import Any, Callable

from pydantic_ai import Agent, RunContext
from pydantic_ai.messages import AgentStreamEvent, ModelMessage
from pydantic_ai.models import Model

from sqlsaber.agents.provider_factory import ProviderFactory
from sqlsaber.config import providers
from sqlsaber.config.settings import Config
from sqlsaber.database import BaseDatabaseConnection
from sqlsaber.database.schema import SchemaManager
from sqlsaber.memory.manager import MemoryManager
from sqlsaber.prompts.claude import SONNET_4_5
from sqlsaber.prompts.dangerous_mode import DANGEROUS_MODE
from sqlsaber.prompts.memory import MEMORY_ADDITION
from sqlsaber.prompts.openai import GPT_5
from sqlsaber.tools.registry import tool_registry
from sqlsaber.tools.sql_tools import SQLTool


class SQLSaberAgent:
    """Pydantic-AI Agent wrapper for SQLSaber with enhanced state management."""

    def __init__(
        self,
        db_connection: BaseDatabaseConnection,
        database_name: str | None = None,
        memory_manager: MemoryManager | None = None,
        thinking_enabled: bool | None = None,
        model_name: str | None = None,
        api_key: str | None = None,
        allow_dangerous: bool = False,
        memory: str | None = None,
    ):
        self.db_connection = db_connection
        self.database_name = database_name
        self.config = Config()
        self.memory_manager = memory_manager or MemoryManager()
        self.memory_override = memory
        self._model_name_override = model_name
        self._api_key_override = api_key
        self.db_type = self.db_connection.display_name
        self.allow_dangerous = allow_dangerous

        self.schema_manager = SchemaManager(self.db_connection)

        self.thinking_enabled = (
            thinking_enabled
            if thinking_enabled is not None
            else self.config.model.thinking_enabled
        )

        self._configure_sql_tools()
        self.agent = self._build_agent()

    def _configure_sql_tools(self) -> None:
        """Ensure SQL tools receive the active database connection and session config."""
        for tool_name in tool_registry.list_tools():
            tool = tool_registry.get_tool(tool_name)
            if isinstance(tool, SQLTool):
                tool.set_connection(self.db_connection, self.schema_manager)
                tool.allow_dangerous = self.allow_dangerous

    def _build_agent(self) -> Agent:
        """Create and configure the pydantic-ai Agent."""
        if self._api_key_override and not self._model_name_override:
            raise ValueError(
                "Model name is required when providing an api_key override."
            )

        model_name = self._model_name_override or self.config.model.name
        model_name_only = (
            model_name.split(":", 1)[1] if ":" in model_name else model_name
        )

        if not (self._model_name_override and self._api_key_override):
            self.config.auth.validate(model_name)

        provider = providers.provider_from_model(model_name) or ""

        api_key = self._api_key_override or self.config.auth.get_api_key(model_name)

        factory = ProviderFactory()
        agent = factory.create_agent(
            provider=provider,
            model_name=model_name_only,
            full_model_str=model_name,
            api_key=api_key,
            thinking_enabled=self.thinking_enabled,
        )

        self._setup_system_prompt(agent)
        self._register_tools(agent)
        return agent

    def _prompt_memory_text(self, include_memory: bool = True) -> str | None:
        if not include_memory:
            return None

        if self.memory_override is not None:
            mem = self.memory_override.strip()
            return mem or None

        if self.database_name:
            mem = self.memory_manager.format_memories_for_prompt(self.database_name)
            mem = mem.strip()
            return mem or None

        return None

    def _setup_system_prompt(self, agent: Agent) -> None:
        """Configure the agent's system prompt using a simple prompt string."""

        @agent.system_prompt(dynamic=True)
        async def sqlsaber_system_prompt(ctx: RunContext) -> str:
            if isinstance(agent.model, Model) and "gpt-5" in agent.model.model_name:
                base = GPT_5.format(db=self.db_type)

                if self.allow_dangerous:
                    base += DANGEROUS_MODE

                mem = self._prompt_memory_text(include_memory=True)
                if mem:
                    return f"{base}\n\n{MEMORY_ADDITION}\n\n{mem}"

                return base

            return self.system_prompt_text(include_memory=True)

    def system_prompt_text(self, include_memory: bool = True) -> str:
        """Return the original SQLSaber system prompt as a single string."""
        base = SONNET_4_5.format(db=self.db_type)

        if self.allow_dangerous:
            base += DANGEROUS_MODE

        mem = self._prompt_memory_text(include_memory=include_memory)
        if mem:
            return f"{base}\n\n{MEMORY_ADDITION}\n\n{mem}\n\n"

        return base

    def _register_tools(self, agent: Agent) -> None:
        """Register all the SQL tools with the agent."""
        for tool_name in tool_registry.list_tools():
            tool = tool_registry.get_tool(tool_name)
            agent.tool_plain(name=tool.name)(tool.execute)

    def set_thinking(self, enabled: bool) -> None:
        """Update thinking settings and rebuild the agent."""
        self.thinking_enabled = enabled
        self.agent = self._build_agent()

    async def run(
        self,
        prompt: str,
        message_history: Sequence[ModelMessage] | None = None,
        event_stream_handler: Callable[
            [RunContext[Any], AsyncIterable[AgentStreamEvent]],
            Awaitable[None],
        ]
        | None = None,
    ) -> Any:
        """Run the agent."""
        return await self.agent.run(
            prompt,
            message_history=message_history,
            event_stream_handler=event_stream_handler,
        )



"""Interactive mode handling for the CLI."""

import asyncio
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import platformdirs
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from sqlsaber.cli.completers import (
    CompositeCompleter,
    SlashCommandCompleter,
    TableNameCompleter,
)
from sqlsaber.cli.display import DisplayManager
from sqlsaber.cli.slash_commands import CommandContext, SlashCommandProcessor
from sqlsaber.cli.streaming import StreamingQueryHandler
from sqlsaber.cli.usage import SessionUsage
from sqlsaber.config.logging import get_logger
from sqlsaber.database import (
    CSVConnection,
    CSVsConnection,
    DuckDBConnection,
    MySQLConnection,
    PostgreSQLConnection,
    SQLiteConnection,
)
from sqlsaber.database.schema import SchemaManager
from sqlsaber.theme.manager import get_theme_manager
from sqlsaber.threads.manager import ThreadManager

if TYPE_CHECKING:
    from sqlsaber.agents.pydantic_ai_agent import SQLSaberAgent


class InteractiveSession:
    """Manages interactive CLI sessions."""

    def __init__(
        self,
        console: Console,
        sqlsaber_agent: "SQLSaberAgent",
        db_conn,
        database_name: str,
        *,
        initial_thread_id: str | None = None,
        initial_history: list | None = None,
    ):
        self.console = console
        self.sqlsaber_agent = sqlsaber_agent
        self.db_conn = db_conn
        self.database_name = database_name
        self.display = DisplayManager(console)
        self.streaming_handler = StreamingQueryHandler(console)
        self.current_task: asyncio.Task | None = None
        self.cancellation_token: asyncio.Event | None = None
        self.table_completer = TableNameCompleter()
        self.message_history: list | None = initial_history or []
        self.tm = get_theme_manager()

        # Component Managers
        self.thread_manager = ThreadManager(initial_thread_id)
        self.command_processor = SlashCommandProcessor()
        self.session_usage = SessionUsage()

        self.log = get_logger(__name__)

    def _history_path(self) -> Path:
        """Get the history file path, ensuring directory exists."""
        history_dir = Path(platformdirs.user_config_dir("sqlsaber"))
        history_dir.mkdir(parents=True, exist_ok=True)
        return history_dir / "history"

    def _bottom_toolbar(self):
        """Get the bottom toolbar text."""
        return [
            (
                "class:bottom-toolbar",
                " Use 'Esc-Enter' or 'Meta-Enter' to submit.",
            )
        ]

    def _banner(self) -> str:
        """Get the ASCII banner."""
        return """[primary]
███████  ██████  ██      ███████  █████  ██████  ███████ ██████
██      ██    ██ ██      ██      ██   ██ ██   ██ ██      ██   ██
███████ ██    ██ ██      ███████ ███████ ██████  █████   ██████
     ██ ██ ▄▄ ██ ██           ██ ██   ██ ██   ██ ██      ██   ██
███████  ██████  ███████ ███████ ██   ██ ██████  ███████ ██   ██
            ▀▀
    [/primary]"""

    def _instructions(self) -> str:
        """Get the instruction text."""
        return dedent("""
                    - Use `/` for slash commands
                    - Type `@` to get table name completions
                    - Start message with `#` to add something to agent's memory
                    - Use `Ctrl+C` to interrupt and `Ctrl+D` to exit
                    """)

    def _db_type_name(self) -> str:
        """Get human-readable database type name."""
        mapping = {
            PostgreSQLConnection: "PostgreSQL",
            MySQLConnection: "MySQL",
            DuckDBConnection: "DuckDB",
            CSVConnection: "DuckDB",
            CSVsConnection: "DuckDB",
            SQLiteConnection: "SQLite",
        }
        for cls, name in mapping.items():
            if isinstance(self.db_conn, cls):
                return name
        return "database"

    def show_welcome_message(self):
        """Display welcome message for interactive mode."""
        if self.thread_manager.first_message:
            self.console.print(Panel.fit(self._banner(), border_style="primary"))
            self.console.print(
                Markdown(
                    self._instructions(),
                    code_theme=self.tm.pygments_style_name,
                    inline_code_theme=self.tm.pygments_style_name,
                )
            )

        db_name = self.database_name or "Unknown"
        model_name = self.sqlsaber_agent.agent.model.model_name
        self.console.print(
            f"[heading]\nConnected to {db_name} ({self._db_type_name()})[/heading]\n"
            f"[heading]Model: {model_name}[/heading]\n"
        )

        if self.thread_manager.current_thread_id:
            self.console.print(
                f"[muted]Resuming thread:[/muted] {self.thread_manager.current_thread_id}\n"
            )

    async def _handle_memory(self, content: str):
        """Handle memory addition command."""
        if not content:
            self.console.print("[warning]Empty memory content after '#'[/warning]\n")
            return

        try:
            mm = self.sqlsaber_agent.memory_manager
            if mm and self.database_name:
                memory = mm.add_memory(self.database_name, content)
                self.console.print(f"[success]✓ Memory added:[/success] {content}")
                self.console.print(f"[muted]Memory ID: {memory.id}[/muted]\n")
            else:
                self.console.print(
                    "[warning]Could not add memory (no database context)[/warning]\n"
                )
        except Exception as exc:
            self.console.print(f"[warning]Could not add memory:[/warning] {exc}\n")

    async def _update_table_cache(self):
        """Update the table completer cache with fresh data."""
        try:
            tables_data = await SchemaManager(self.db_conn).list_tables()

            # Parse the table information
            table_list = []
            if isinstance(tables_data, dict) and "tables" in tables_data:
                for table in tables_data["tables"]:
                    if isinstance(table, dict):
                        name = table.get("name", "")
                        schema = table.get("schema", "")
                        full_name = table.get("full_name", "")

                        # Use full_name if available, otherwise construct it
                        if full_name:
                            table_name = full_name
                        elif schema and schema != "main":
                            table_name = f"{schema}.{name}"
                        else:
                            table_name = name

                        # No description needed - cleaner completions
                        table_list.append((table_name, ""))

            # Update the completer cache
            self.table_completer.update_cache(table_list)

        except Exception:
            # If there's an error, just use empty cache
            self.table_completer.update_cache([])

    async def before_prompt_loop(self):
        """Hook to refresh context before prompt loop."""
        await self._update_table_cache()

    async def _execute_query_with_cancellation(self, user_query: str):
        """Execute a query with cancellation support."""
        self.log.info("interactive.query.start", database=self.database_name)
        # Create cancellation token
        self.cancellation_token = asyncio.Event()

        # Create the query task
        query_task = asyncio.create_task(
            self.streaming_handler.execute_streaming_query(
                user_query,
                self.sqlsaber_agent,
                self.cancellation_token,
                self.message_history,
            )
        )
        self.current_task = query_task

        try:
            run_result = await query_task
            # Persist message history from this run using pydantic-ai API
            if run_result is not None:
                self.message_history = await self.thread_manager.save_run(
                    run_result=run_result,
                    database_name=self.database_name,
                    user_query=user_query,
                    model_name=self.sqlsaber_agent.agent.model.model_name,
                )
                # Track usage for session summary
                # Use result.response.usage for the FINAL request's context size
                final_context = run_result.response.usage.input_tokens
                self.session_usage.add_run(run_result.usage(), final_context)
        finally:
            self.current_task = None
            self.cancellation_token = None
            self.log.info("interactive.query.end")

    async def run(self):
        """Run the interactive session loop."""
        self.log.info("interactive.start", database=self.database_name)
        self.show_welcome_message()
        await self.before_prompt_loop()

        session = PromptSession(history=FileHistory(self._history_path()))

        def clear_history():
            self.message_history = []

        while True:
            try:
                with patch_stdout():
                    user_query = await session.prompt_async(
                        "> ",
                        multiline=True,
                        completer=CompositeCompleter(
                            SlashCommandCompleter(), self.table_completer
                        ),
                        bottom_toolbar=self._bottom_toolbar,
                        style=self.tm.pt_style(),
                    )

                user_query = user_query.strip()

                if not user_query:
                    continue

                # Process slash commands
                context = CommandContext(
                    console=self.console,
                    agent=self.sqlsaber_agent,
                    thread_manager=self.thread_manager,
                    on_clear_history=clear_history,
                    session_usage=self.session_usage,
                )

                cmd_result = await self.command_processor.process(user_query, context)
                if cmd_result.should_exit:
                    break
                if cmd_result.handled:
                    continue

                # Handle memory addition
                if user_query.strip().startswith("#"):
                    await self._handle_memory(user_query[1:].strip())
                    continue

                # Execute query with cancellation support
                await self._execute_query_with_cancellation(user_query)
                self.display.show_newline()

            except KeyboardInterrupt:
                # Handle Ctrl+C - cancel current task if running
                if self.current_task and not self.current_task.done():
                    if self.cancellation_token is not None:
                        self.cancellation_token.set()
                    self.current_task.cancel()
                    try:
                        await self.current_task
                    except asyncio.CancelledError:
                        pass
                    self.console.print("\n[warning]Query interrupted[/warning]")
                else:
                    self.console.print(
                        "\n[warning]Press Ctrl+D to exit. Or use '/exit' or '/quit' slash command.[/warning]"
                    )
            except EOFError:
                # Exit when Ctrl+D is pressed
                self.display.show_session_summary(self.session_usage)
                ended_thread_id = await self.thread_manager.end_current_thread()
                if ended_thread_id:
                    hint = f"saber threads resume {ended_thread_id}"
                    self.console.print(
                        f"[muted]You can continue this thread using:[/muted] {hint}"
                    )
                break
            except Exception as exc:
                self.console.print(f"[error]Error:[/error] {exc}")
                self.log.exception("interactive.error", error=str(exc))



"""Database connection resolution from CLI input."""

from dataclasses import dataclass
from pathlib import Path
from urllib.parse import urlencode, urlparse

from sqlsaber.config.database import DatabaseConfig, DatabaseConfigManager


class DatabaseResolutionError(Exception):
    """Exception raised when database resolution fails."""


@dataclass
class ResolvedDatabase:
    """Result of database resolution containing canonical connection info."""

    name: str  # Human-readable name for display/logging
    connection_string: str  # Canonical connection string for DatabaseConnection factory
    excluded_schemas: list[str]


SUPPORTED_SCHEMES = {"postgresql", "mysql", "sqlite", "duckdb", "csv", "csvs"}


def _is_connection_string(s: str) -> bool:
    """Check if string looks like a connection string with supported scheme."""
    try:
        scheme = urlparse(s).scheme
        return scheme in SUPPORTED_SCHEMES
    except Exception:
        return False


def _csv_stem_from_connection_string(connection_string: str) -> str:
    parsed = urlparse(connection_string)
    stem = Path(parsed.path).stem
    return stem or "csv"


def _resolve_multiple_csvs(specs: list[str]) -> ResolvedDatabase:
    csv_specs: list[str] = []
    stems: list[str] = []

    for spec in specs:
        if _is_connection_string(spec):
            scheme = urlparse(spec).scheme
            if scheme != "csv":
                raise DatabaseResolutionError(
                    "Multiple database arguments are only supported for CSV files. "
                    f"Got connection string with scheme '{scheme}': {spec}"
                )
            csv_specs.append(spec)
            stems.append(_csv_stem_from_connection_string(spec))
            continue

        path = Path(spec).expanduser().resolve()
        if path.suffix.lower() != ".csv":
            raise DatabaseResolutionError(
                "Multiple database arguments are only supported for CSV files. "
                f"Got non-CSV path: {spec}"
            )
        if not path.exists():
            raise DatabaseResolutionError(f"CSV file '{spec}' not found.")
        csv_specs.append(f"csv:///{path}")
        stems.append(path.stem or "csv")

    if not csv_specs:
        raise DatabaseResolutionError(
            "Multiple database arguments were provided, but no CSV files were found."
        )

    query = urlencode({"spec": csv_specs}, doseq=True)

    if len(stems) <= 3:
        name = " + ".join(stems)
    else:
        name = f"{stems[0]} + {len(stems) - 1} more"

    return ResolvedDatabase(
        name=name, connection_string=f"csvs:///?{query}", excluded_schemas=[]
    )


def resolve_database(
    spec: str | list[str] | None, config_mgr: DatabaseConfigManager
) -> ResolvedDatabase:
    """Turn user CLI input into resolved database connection info.

    Args:
        spec: User input - None (default), configured name, connection string, file path,
            or a list of CSV file paths/CSV connection strings.
        config_mgr: Database configuration manager for looking up configured connections

    Returns:
        ResolvedDatabase with name and canonical connection string

    Raises:
        DatabaseResolutionError: If the spec cannot be resolved to a valid database connection
    """
    if spec is None:
        db_cfg = config_mgr.get_default_database()
        if not db_cfg:
            raise DatabaseResolutionError(
                "No database connections configured. "
                "Use 'sqlsaber db add <name>' to add one."
            )
        return ResolvedDatabase(
            name=db_cfg.name,
            connection_string=db_cfg.to_connection_string(),
            excluded_schemas=list(db_cfg.exclude_schemas),
        )

    if isinstance(spec, list):
        if len(spec) == 1:
            return resolve_database(spec[0], config_mgr)
        if len(spec) > 1:
            return _resolve_multiple_csvs(spec)
        raise DatabaseResolutionError("Empty database argument list.")

    # 1. Connection string?
    if _is_connection_string(spec):
        scheme = urlparse(spec).scheme
        if scheme in {"postgresql", "mysql"}:
            db_name = urlparse(spec).path.lstrip("/") or "database"
        elif scheme in {"sqlite", "duckdb", "csv", "csvs"}:
            db_name = Path(urlparse(spec).path).stem or "database"
        else:  # should not happen because of SUPPORTED_SCHEMES
            db_name = "database"
        return ResolvedDatabase(
            name=db_name, connection_string=spec, excluded_schemas=[]
        )

    # 2. Raw file path?
    path = Path(spec).expanduser().resolve()
    if path.suffix.lower() == ".csv":
        if not path.exists():
            raise DatabaseResolutionError(f"CSV file '{spec}' not found.")
        return ResolvedDatabase(
            name=path.stem, connection_string=f"csv:///{path}", excluded_schemas=[]
        )
    if path.suffix.lower() in {".db", ".sqlite", ".sqlite3"}:
        if not path.exists():
            raise DatabaseResolutionError(f"SQLite file '{spec}' not found.")
        return ResolvedDatabase(
            name=path.stem, connection_string=f"sqlite:///{path}", excluded_schemas=[]
        )
    if path.suffix.lower() in {".duckdb", ".ddb"}:
        if not path.exists():
            raise DatabaseResolutionError(f"DuckDB file '{spec}' not found.")
        return ResolvedDatabase(
            name=path.stem, connection_string=f"duckdb:///{path}", excluded_schemas=[]
        )

    # 3. Must be a configured name
    db_cfg: DatabaseConfig | None = config_mgr.get_database(spec)
    if not db_cfg:
        raise DatabaseResolutionError(
            f"Database connection '{spec}' not found. "
            "Use 'sqlsaber db list' to see available connections."
        )
    return ResolvedDatabase(
        name=db_cfg.name,
        connection_string=db_cfg.to_connection_string(),
        excluded_schemas=list(db_cfg.exclude_schemas),
    )



DANGEROUS_MODE = """

## Write Operations (Dangerous Mode)

You are operating in DANGEROUS MODE. This means:

- You MAY generate and execute INSERT, UPDATE, DELETE statements when the user explicitly asks to modify data.
- You MAY generate and execute CREATE and ALTER statements when the user explicitly asks to modify schema.
- UPDATE and DELETE statements MUST include a WHERE clause; unfiltered mutations are blocked.
- DROP and TRUNCATE statements are NEVER executed by this tool. If a user asks to drop or truncate tables, show them the SQL but tell them to run it manually.
- Prefer minimal, targeted changes; never run broad operations without filters unless explicitly requested.
- Always explain what changes will be made before executing write operations.
"""



SONNET_4_5 = """You are a helpful SQL assistant designed to help users query their {db} database using natural language requests.

## Your Core Responsibilities

1. **Understand requests**: Convert natural language queries into appropriate SQL
2. **Explore schema**: Use provided tools to discover and understand database structure
3. **Generate queries**: Create safe, efficient SQL queries with proper syntax
4. **Execute safely**: Run only SELECT queries (no data modification allowed)
5. **Explain results**: Format outputs clearly and explain what the queries accomplish

## Tool Usage Strategy

You must follow this systematic approach:

1. **Always start with `list_tables`** to discover available tables
2. **Use `introspect_schema` strategically** with table patterns (like 'sample%' or '%experiment%') to get details only for relevant tables
3. **Execute SQL queries safely** - all SELECT statements automatically include LIMIT clauses for safety

## Important Guidelines

- Write proper JOIN syntax and avoid cartesian products
- Include appropriate WHERE clauses to filter results effectively
- Convert timestamp columns to text format in your queries
- Use parameterized queries when needed for security
- Handle errors gracefully and suggest fixes when issues arise
- Explain each query's purpose in simple non-technical terms

## Response Format

For each user request, structure your response as follows:

Before proceeding with database exploration, work through the problem systematically in <analysis> tags inside your thinking block:
- Parse the user's natural language request and identify the core question being asked
- Extract key terms, entities, and concepts that might correspond to database tables or columns
- Consider what types of data relationships might be involved (e.g., one-to-many, many-to-many)
- Plan your database exploration approach step by step
- Design your overall SQL strategy, including potential JOINs, filters, and aggregations
- Anticipate potential challenges or edge cases specific to this database type
- Verify your approach makes logical sense for the business question

It's OK for this analysis section to be quite long if the request is complex.

Then, execute the planned database exploration and queries, providing clear explanations of results.

## Example Response Structure

Working through this systematically in my analysis, then exploring tables and executing queries...

Now I need to address your specific request. Before proceeding with database exploration, let me analyze what you're asking for:

Your final response should focus on the database exploration, query execution, and results explanation, without duplicating or rehashing the analytical work done in the thinking block.
"""



"""Check for available updates on PyPI.

This is best-effort and must never fail the CLI.
"""

import asyncio
import os
from dataclasses import dataclass
from importlib import metadata

import httpx
from rich.console import Console

from sqlsaber.config.logging import get_logger
from sqlsaber.theme.manager import create_console

PACKAGE_NAME = "sqlsaber"
ENV_SKIP_VERSION_CHECK = "SQLSABER_SKIP_VERSION_CHECK"
PYPI_URL = f"https://pypi.org/pypi/{PACKAGE_NAME}/json"

_LOG = get_logger(__name__)
_SCHEDULED = False


@dataclass(frozen=True)
class VersionInfo:
    current: str
    latest: str


def _env_disables_check() -> bool:
    value = os.getenv(ENV_SKIP_VERSION_CHECK)
    if value is None:
        return False

    normalized = value.strip().lower()
    if normalized in {"0", "false", "no", "off"}:
        return False
    return True


def _get_version_from_pyproject() -> str | None:
    try:
        import tomllib
        from pathlib import Path

        root = Path(__file__).resolve().parents[3]
        pyproject = root / "pyproject.toml"
        if not pyproject.is_file():
            return None

        data = tomllib.loads(pyproject.read_text(encoding="utf-8"))
        version = data.get("project", {}).get("version")
        return version if isinstance(version, str) else None
    except Exception:
        return None


def _get_current_version() -> str | None:
    try:
        return metadata.version(PACKAGE_NAME)
    except metadata.PackageNotFoundError:
        return _get_version_from_pyproject()
    except Exception:
        return None


async def _fetch_latest_version() -> str | None:
    timeout = httpx.Timeout(connect=1.0, read=1.5, write=1.0, pool=1.0)

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            resp = await client.get(PYPI_URL)
            resp.raise_for_status()
            data = resp.json()
            version = data.get("info", {}).get("version")
            return version if isinstance(version, str) else None
    except Exception:
        return None


def _parse_version(v: str) -> tuple[int, ...]:
    parts: list[int] = []
    for part in v.split("."):
        num = ""
        for ch in part:
            if ch.isdigit():
                num += ch
            else:
                break
        parts.append(int(num) if num else 0)
    return tuple(parts)


def _is_newer(latest: str, current: str) -> bool:
    try:
        from packaging.version import Version

        return Version(latest) > Version(current)
    except Exception:
        return _parse_version(latest) > _parse_version(current)


def _print_update_notice(console: Console, info: VersionInfo) -> None:
    console.print(f"\nUpdate available: {info.current} → {info.latest}\n")
    console.print(f"Run: uv tool update {PACKAGE_NAME}\n")


async def _check_and_notify(console: Console) -> None:
    current = _get_current_version()
    if not current:
        return

    latest = await _fetch_latest_version()
    if not latest:
        return

    if _is_newer(latest, current):
        _LOG.info("update_check.available", current=current, latest=latest)
        _print_update_notice(console, VersionInfo(current=current, latest=latest))


async def _run_safely(console: Console) -> None:
    try:
        await _check_and_notify(console)
    except Exception:
        return


def schedule_update_check(console: Console | None = None) -> None:
    """Schedule a non-blocking update check.

    Safe to call multiple times; only schedules once per process.
    """

    global _SCHEDULED
    if _SCHEDULED:
        return

    if _env_disables_check():
        return

    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return

    _SCHEDULED = True
    asyncio.create_task(_run_safely(console or create_console()))



"""CLI command definitions and handlers."""

import asyncio
import sys
from typing import Annotated

import cyclopts
from rich.panel import Panel

from sqlsaber.cli.auth import create_auth_app
from sqlsaber.cli.database import create_db_app
from sqlsaber.cli.memory import create_memory_app
from sqlsaber.cli.models import create_models_app
from sqlsaber.cli.onboarding import needs_onboarding, run_onboarding
from sqlsaber.cli.theme import create_theme_app
from sqlsaber.cli.threads import create_threads_app
from sqlsaber.cli.update_check import schedule_update_check

# Lazy imports - only import what's needed for CLI parsing
from sqlsaber.config.database import DatabaseConfigManager
from sqlsaber.config.logging import get_logger, setup_logging
from sqlsaber.theme.manager import create_console

DANGEROUS_MODE_WARNING = (
    "The assistant can execute INSERT/UPDATE/DELETE and DDL statements."
)


class CLIError(Exception):
    """Exception raised for CLI errors that should result in exit."""

    def __init__(self, message: str, exit_code: int = 1):
        super().__init__(message)
        self.exit_code = exit_code


app = cyclopts.App(
    name="sqlsaber",
    help="SQLsaber - Open-source agentic SQL assistant for your database",
)

app.command(create_auth_app(), name="auth")
app.command(create_db_app(), name="db")
app.command(create_memory_app(), name="memory")
app.command(create_models_app(), name="models")
app.command(create_theme_app(), name="theme")
app.command(create_threads_app(), name="threads")

console = create_console()
config_manager = DatabaseConfigManager()


@app.meta.default
def meta_handler(
    database: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            ["--database", "-d"],
            help="Database connection name, file path (CSV/SQLite/DuckDB), connection string (postgresql://, mysql://, duckdb://), or one/more CSV files via repeated -d (uses default if not specified)",
        ),
    ] = None,
):
    """
    Query your database using natural language.

    Examples:
        saber                                  # Start interactive mode
        saber "show me all users"              # Run a single query with default database
        saber -d mydb "show me users"          # Run a query with specific database
        saber -d data.csv "show me users"      # Run a query with ad-hoc CSV file
        saber -d users.csv -d orders.csv "join users and orders"  # Multiple CSV files (one view per file)
        saber -d data.db "show me users"       # Run a query with ad-hoc SQLite file
        saber -d data.duckdb "show me users"   # Run a query with ad-hoc DuckDB file
        saber -d "postgresql://user:pass@host:5432/db" "show users"  # PostgreSQL connection string
        saber -d "mysql://user:pass@host:3306/db" "show users"       # MySQL connection string
        saber -d "duckdb:///data.duckdb" "show users"                 # DuckDB connection string
        echo "show me all users" | saber       # Read query from stdin
        cat query.txt | saber                  # Read query from file via stdin
    """


@app.default
def query(
    query_text: Annotated[
        str | None,
        cyclopts.Parameter(
            help="Question in natural language (if not provided, reads from stdin or starts interactive mode)",
        ),
    ] = None,
    database: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            ["--database", "-d"],
            help="Database connection name, file path (CSV/SQLite/DuckDB), connection string (postgresql://, mysql://, duckdb://), or one/more CSV files via repeated -d (uses default if not specified)",
        ),
    ] = None,
    thinking: bool = False,
    allow_dangerous: Annotated[
        bool,
        cyclopts.Parameter(
            ["--allow-dangerous"],
            help="Allow INSERT/UPDATE/DELETE/DDL statements (DROP/TRUNCATE always blocked; UPDATE/DELETE require WHERE clause)",
        ),
    ] = False,
):
    """Run a query against the database or start interactive mode.

    When called without arguments:
    - If stdin has data, reads query from stdin
    - Otherwise, starts interactive mode

    When called with a query string, executes that query and exits.

    Examples:
        saber                             # Start interactive mode
        saber "show me all users"         # Run a single query
        saber -d data.csv "show users"    # Run a query with ad-hoc CSV file
        saber -d users.csv -d orders.csv "join users and orders"  # Multiple CSV files (one view per file)
        saber -d data.db "show users"     # Run a query with ad-hoc SQLite file
        saber -d data.duckdb "show users" # Run a query with ad-hoc DuckDB file
        saber -d "postgresql://user:pass@host:5432/db" "show users"  # PostgreSQL connection string
        saber -d "mysql://user:pass@host:3306/db" "show users"       # MySQL connection string
        saber -d "duckdb:///data.duckdb" "show users"                 # DuckDB connection string
        echo "show me all users" | saber  # Read query from stdin
    """

    async def run_session():
        schedule_update_check(console)

        log = get_logger(__name__)
        log.info(
            "cli.session.start",
            argv=sys.argv[1:],
            database=database,
            has_query=query_text is not None,
            thinking=thinking,
            allow_dangerous=allow_dangerous,
        )
        # Import heavy dependencies only when actually running a query
        # This is only done to speed up startup time
        from sqlsaber.agents import SQLSaberAgent
        from sqlsaber.cli.display import DisplayManager
        from sqlsaber.cli.interactive import InteractiveSession
        from sqlsaber.cli.streaming import StreamingQueryHandler
        from sqlsaber.cli.usage import SessionUsage
        from sqlsaber.database import (
            DatabaseConnection,
        )
        from sqlsaber.database.resolver import DatabaseResolutionError, resolve_database
        from sqlsaber.threads import ThreadStorage

        # Check if query_text is None and stdin has data
        actual_query = query_text
        if query_text is None and not sys.stdin.isatty():
            # Read from stdin
            actual_query = sys.stdin.read().strip()
            if not actual_query:
                # If stdin was empty, fall back to interactive mode
                actual_query = None

        # Check if onboarding is needed (only for interactive mode or when no database is configured)
        if needs_onboarding(database):
            # Run onboarding flow
            log.debug("cli.onboarding.start")
            onboarding_success = await run_onboarding()
            if not onboarding_success:
                # User cancelled or onboarding failed
                raise CLIError(
                    "Setup incomplete. Please configure your database and try again."
                )
            log.info("cli.onboarding.complete", success=True)

        # Resolve database from CLI input
        try:
            resolved = resolve_database(database, config_manager)
            connection_string = resolved.connection_string
            db_name = resolved.name
            log.info(
                "db.resolve.success",
                name=db_name,
            )
        except DatabaseResolutionError as e:
            log.error("db.resolve.error", error=str(e))
            raise CLIError(str(e))

        # Create database connection
        try:
            db_conn = DatabaseConnection(
                connection_string, excluded_schemas=resolved.excluded_schemas
            )
            log.info("db.connection.created", db_type=type(db_conn).__name__)
        except Exception as e:
            log.exception("db.connection.error", error=str(e))
            raise CLIError(f"Error creating database connection: {e}")

        # Create pydantic-ai agent instance with database name for memory context
        sqlsaber_agent = SQLSaberAgent(
            db_conn,
            db_name,
            thinking_enabled=thinking,
            allow_dangerous=allow_dangerous,
        )

        try:
            if actual_query:
                # Single query mode with streaming
                streaming_handler = StreamingQueryHandler(console)
                db_type = sqlsaber_agent.db_type
                model_name = sqlsaber_agent.config.model.name
                console.print(
                    f"[primary]Connected to:[/primary] {db_name} ({db_type})\n"
                    f"[primary]Model:[/primary] {model_name}\n"
                )
                if allow_dangerous:
                    console.print(
                        Panel(
                            DANGEROUS_MODE_WARNING,
                            title=":warning: DANGEROUS MODE ENABLED",
                            style="warning",
                        )
                    )
                log.info("query.execute.start", db_name=db_name, db_type=db_type)
                run = await streaming_handler.execute_streaming_query(
                    actual_query, sqlsaber_agent
                )

                # Track and display session usage
                if run is not None:
                    session_usage = SessionUsage()
                    final_context = run.response.usage.input_tokens
                    session_usage.add_run(run.usage(), final_context)
                    display = DisplayManager(console)
                    display.show_session_summary(session_usage)

                # Persist non-interactive run as a thread snapshot so it can be resumed later
                threads: ThreadStorage | None = None
                try:
                    if run is not None:
                        threads = ThreadStorage()

                        thread_id = await threads.save_snapshot(
                            messages_json=run.all_messages_json(),
                            database_name=db_name,
                        )
                        await threads.save_metadata(
                            thread_id=thread_id,
                            title=actual_query,
                            model_name=sqlsaber_agent.config.model.name,
                        )
                        await threads.end_thread(thread_id)
                        console.print(
                            f"[dim]You can continue this thread using:[/dim] saber threads resume {thread_id}"
                        )
                        log.info("thread.save.success", thread_id=thread_id)
                except Exception:
                    # best-effort persistence; don't fail the CLI on storage errors
                    log.warning("thread.save.failed", exc_info=True)
                    pass
                finally:
                    if threads is not None:
                        await threads.prune_threads()
            else:
                # Interactive mode
                if allow_dangerous:
                    console.print(
                        Panel(
                            DANGEROUS_MODE_WARNING,
                            title=":warning: DANGEROUS MODE ENABLED",
                            style="warning",
                        )
                    )
                session = InteractiveSession(console, sqlsaber_agent, db_conn, db_name)
                await session.run()

        finally:
            # Clean up
            await db_conn.close()
            log.info("db.connection.closed")
            console.print("\n[success]Goodbye![/success]")

    # Run the async function with proper error handling
    try:
        asyncio.run(run_session())
    except CLIError as e:
        get_logger(__name__).error("cli.error", error=str(e))
        console.print(f"[error]Error:[/error] {e}")
        sys.exit(e.exit_code)


def main():
    """Entry point for the CLI application."""
    setup_logging()
    get_logger(__name__).info("cli.start")
    app()



"""Streaming query handling for the CLI (pydantic-ai based).

This module uses DisplayManager's LiveMarkdownRenderer to stream Markdown
incrementally as the agent outputs tokens. Tool calls and results are
rendered via DisplayManager helpers.
"""

import asyncio
import json
from functools import singledispatchmethod
from typing import TYPE_CHECKING, AsyncIterable

from pydantic_ai import RunContext
from pydantic_ai.messages import (
    AgentStreamEvent,
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    PartDeltaEvent,
    PartStartEvent,
    TextPart,
    TextPartDelta,
    ThinkingPart,
    ThinkingPartDelta,
)
from rich.console import Console

from sqlsaber.cli.display import DisplayManager
from sqlsaber.config.logging import get_logger

if TYPE_CHECKING:
    from sqlsaber.agents.pydantic_ai_agent import SQLSaberAgent


class StreamingQueryHandler:
    """
    Handles streaming query execution and display using pydantic-ai events.

    Uses DisplayManager.live to render Markdown incrementally as text streams in.
    """

    def __init__(self, console: Console):
        self.console = console
        self.display = DisplayManager(console)
        self.log = get_logger(__name__)

    async def _event_stream_handler(
        self, ctx: RunContext, event_stream: AsyncIterable[AgentStreamEvent]
    ) -> None:
        """
        Handle pydantic-ai streaming events and update Live Markdown via DisplayManager.
        """

        async for event in event_stream:
            await self.on_event(event, ctx)

    # --- Event routing via singledispatchmethod ---------------------------------------
    @singledispatchmethod
    async def on_event(
        self, event: AgentStreamEvent, ctx: RunContext
    ) -> None:  # default
        return

    @on_event.register
    async def _(self, event: PartStartEvent, ctx: RunContext) -> None:
        if isinstance(event.part, TextPart):
            self.display.live.ensure_segment(TextPart)
            self.display.live.append(event.part.content)
        elif isinstance(event.part, ThinkingPart):
            self.display.live.ensure_segment(ThinkingPart)
            self.display.live.append(event.part.content)

    @on_event.register
    async def _(self, event: PartDeltaEvent, ctx: RunContext) -> None:
        d = event.delta
        if isinstance(d, TextPartDelta):
            delta = d.content_delta or ""
            if delta:
                self.display.live.ensure_segment(TextPart)
                self.display.live.append(delta)
        elif isinstance(d, ThinkingPartDelta):
            delta = d.content_delta or ""
            if delta:
                self.display.live.ensure_segment(ThinkingPart)
                self.display.live.append(delta)

    @on_event.register
    async def _(self, event: FunctionToolCallEvent, ctx: RunContext) -> None:
        # Clear any status/markdown Live so tool output sits between
        self.display.live.end_status()
        self.display.live.end_if_active()
        args = event.part.args_as_dict()

        # Special handling: display SQL via Live as markdown code block
        if event.part.tool_name == "execute_sql":
            query = args.get("query") or ""
            if isinstance(query, str) and query.strip():
                self.display.live.start_sql_block(query)
        else:
            self.display.show_tool_executing(event.part.tool_name, args)

    @on_event.register
    async def _(self, event: FunctionToolResultEvent, ctx: RunContext) -> None:
        # Route tool result to appropriate display
        tool_name = event.result.tool_name
        content = event.result.content
        if tool_name == "list_tables":
            if isinstance(content, (str, dict)):
                self.display.show_table_list(content)
        elif tool_name == "introspect_schema":
            if isinstance(content, (str, dict)):
                self.display.show_schema_info(content)
        elif tool_name == "execute_sql":
            data: dict[str, object] = {}
            if isinstance(content, str):
                try:
                    data = json.loads(content)
                except (json.JSONDecodeError, TypeError) as exc:
                    try:
                        self.console.log(f"Malformed execute_sql result: {exc}")
                    except Exception:
                        pass
            elif isinstance(content, dict):
                data = content

            if isinstance(data, dict):
                if data.get("success") and data.get("results"):
                    self.display.show_query_results(data["results"])  # type: ignore[arg-type]
                elif "error" in data:
                    error_msg = data.get("error")
                    if isinstance(error_msg, str):
                        self.display.show_sql_error(error_msg, data.get("suggestions"))  # type: ignore[arg-type]
        # Add a blank line after tool output to separate from next segment
        self.display.show_newline()
        # Show status while agent sends a follow-up request to the model
        self.display.live.start_status("Crunching data...")

    async def execute_streaming_query(
        self,
        user_query: str,
        sqlsaber_agent: "SQLSaberAgent",
        cancellation_token: asyncio.Event | None = None,
        message_history: list | None = None,
    ):
        self.display.live.prepare_code_blocks()
        try:
            self.log.info("streaming.execute.start")
            self.display.live.start_status("Crunching data...")

            run = await sqlsaber_agent.run(
                user_query,
                message_history=message_history,
                event_stream_handler=self._event_stream_handler,
            )
            self.log.info("streaming.execute.end")
            return run
        except asyncio.CancelledError:
            self.display.show_newline()
            self.console.print("[warning]Query interrupted[/warning]")
            self.log.info("streaming.execute.cancelled")
            return None
        finally:
            try:
                self.display.live.end_status()
            finally:
                self.display.live.end_if_active()



"""Interactive onboarding flow for first-time SQLSaber users."""

import sys

from rich.panel import Panel

from sqlsaber.cli.models import ModelManager
from sqlsaber.config.api_keys import APIKeyManager
from sqlsaber.config.auth import AuthConfigManager
from sqlsaber.config.database import DatabaseConfigManager
from sqlsaber.theme.manager import create_console

console = create_console()


def needs_onboarding(database_arg: str | list[str] | None = None) -> bool:
    """Check if user needs onboarding.

    Onboarding is needed if:
    - No database is configured AND no database connection string provided via CLI
    """
    # If user provided a database argument, skip onboarding
    if database_arg:
        return False

    # Check if databases are configured
    db_manager = DatabaseConfigManager()
    has_db = db_manager.has_databases()

    return not has_db


def welcome_screen() -> None:
    """Display welcome screen to new users."""
    banner = """[primary]
███████  ██████  ██      ███████  █████  ██████  ███████ ██████
██      ██    ██ ██      ██      ██   ██ ██   ██ ██      ██   ██
███████ ██    ██ ██      ███████ ███████ ██████  █████   ██████
     ██ ██ ▄▄ ██ ██           ██ ██   ██ ██   ██ ██      ██   ██
███████  ██████  ███████ ███████ ██   ██ ██████  ███████ ██   ██
           ▀▀
    [/primary]"""

    console.print(Panel.fit(banner, style="primary"))
    console.print()

    welcome_message = """
[bold]Welcome to SQLsaber! 🎉[/bold]

SQLsaber is an agentic SQL assistant that lets you query your database using natural language.

Let's get you set up in just a few steps.
    """

    console.print(
        Panel.fit(welcome_message.strip(), border_style="primary", padding=(1, 2))
    )
    console.print()


async def setup_database_guided() -> str | None:
    """Guide user through database setup.

    Returns the name of the configured database or None if cancelled.
    """
    from sqlsaber.application.db_setup import (
        build_config,
        collect_db_input,
        save_database,
        test_connection,
    )
    from sqlsaber.application.prompts import AsyncPrompter

    console.print("[heading]Step 1 of 2: Database Connection[/heading]")
    console.print()

    try:
        # Ask for connection name
        prompter = AsyncPrompter()
        name = await prompter.text(
            "What would you like to name this connection?",
            default="mydb",
            validate=lambda x: bool(x.strip()) or "Name cannot be empty",
        )

        if name is None:
            return None

        name = name.strip()

        # Check if name already exists
        db_manager = DatabaseConfigManager()
        if db_manager.get_database(name):
            console.print(
                f"[warning]Database connection '{name}' already exists.[/warning]"
            )
            return name

        # Collect database input (simplified - no SSL in onboarding)
        db_input = await collect_db_input(
            prompter=prompter, name=name, db_type="postgresql", include_ssl=False
        )

        if db_input is None:
            return None

        # Build config
        db_config = build_config(db_input)

        # Test the connection
        console.print(f"[muted]Testing connection to '{name}'...[/muted]")
        connection_success = await test_connection(db_config, db_input.password)

        if not connection_success:
            retry = await prompter.confirm(
                "Would you like to try again with different settings?", default=True
            )
            if retry:
                return await setup_database_guided()
            else:
                console.print(
                    "[warning]You can add a database later using 'saber db add'[/warning]"
                )
                return None

        # Save the configuration
        try:
            save_database(db_manager, db_config, db_input.password)
            console.print(f"[success]✓ Connection to '{name}' successful![/success]")
            console.print()
            return name
        except Exception as e:
            console.print(f"[error]Error saving database:[/error] {e}")
            return None

    except KeyboardInterrupt:
        console.print("\n[warning]Setup cancelled.[/warning]")
        return None
    except Exception as e:
        console.print(f"[error]Unexpected error:[/error] {e}")
        return None


async def select_model_for_provider(provider: str) -> str | None:
    """Fetch and let user select a model for the given provider.

    Returns the selected model ID or None if cancelled/failed.
    """
    from sqlsaber.application.model_selection import choose_model, fetch_models
    from sqlsaber.application.prompts import AsyncPrompter

    try:
        console.print()
        console.print(f"[muted]Fetching available {provider.title()} models...[/muted]")

        model_manager = ModelManager()
        models = await fetch_models(model_manager, providers=[provider])

        if not models:
            console.print(
                f"[warning]Could not fetch models for {provider}. Using default.[/warning]"
            )
            # Use provider-specific default or fallback to Anthropic
            default_model_id = ModelManager.RECOMMENDED_MODELS.get(
                provider, ModelManager.DEFAULT_MODEL
            )
            # Format it properly if we have a recommended model for this provider
            if provider in ModelManager.RECOMMENDED_MODELS:
                return f"{provider}:{ModelManager.RECOMMENDED_MODELS[provider]}"
            return default_model_id

        prompter = AsyncPrompter()
        console.print()
        selected_model = await choose_model(
            prompter, models, restrict_provider=provider, use_search_filter=True
        )

        return selected_model

    except KeyboardInterrupt:
        console.print("\n[warning]Model selection cancelled.[/warning]")
        return None
    except Exception as e:
        console.print(f"[warning]Error selecting model: {e}. Using default.[/warning]")
        # Fallback to provider default
        if provider in ModelManager.RECOMMENDED_MODELS:
            return f"{provider}:{ModelManager.RECOMMENDED_MODELS[provider]}"
        return ModelManager.DEFAULT_MODEL


async def setup_auth_guided() -> tuple[bool, str | None]:
    """Guide user through auth setup.

    Returns tuple of (success: bool, selected_model: str | None).
    """
    from sqlsaber.application.auth_setup import setup_auth
    from sqlsaber.application.prompts import AsyncPrompter

    console.print("[primary]Step 2 of 2: Authentication[/primary]")
    console.print()

    try:
        # Run auth setup
        prompter = AsyncPrompter()
        auth_manager = AuthConfigManager()
        api_key_manager = APIKeyManager()

        success, provider = await setup_auth(
            prompter=prompter,
            auth_manager=auth_manager,
            api_key_manager=api_key_manager,
            default_provider="anthropic",
        )

        if not success:
            console.print(
                "[warning]You can set it up later using 'saber auth setup'[/warning]"
            )
            console.print()
            return False, None

        # If auth configured but we don't know the provider (already configured case)
        if provider is None:
            console.print()
            return True, None

        # Select model for this provider
        selected_model = await select_model_for_provider(provider)
        if selected_model:
            model_manager = ModelManager()
            model_manager.set_model(selected_model)
            console.print(f"[success]✓ Model set to: {selected_model}[/success]")
        console.print()
        return True, selected_model

    except KeyboardInterrupt:
        console.print("\n[warning]Setup cancelled.[/warning]")
        console.print()
        return False, None
    except Exception as e:
        console.print(f"[error]Unexpected error:[/error] {e}")
        console.print()
        return False, None


def success_screen(
    database_name: str | None, auth_configured: bool, model_name: str | None = None
) -> None:
    """Display success screen after onboarding."""

    console.print("[success]You're all set! 🚀[/success]")
    console.print()

    if database_name and auth_configured:
        console.print(
            f"[success]✓ Database '{database_name}' connected and ready to use[/success]"
        )
        console.print("[success]✓ Authentication configured[/success]")
        if model_name:
            console.print(f"[success]✓ Model: {model_name}[/success]")
    elif database_name:
        console.print(
            f"[success]✓ Database '{database_name}' connected and ready to use[/success]"
        )
        console.print(
            "[warning]⚠ AI authentication not configured - you'll be prompted when needed[/warning]"
        )
    elif auth_configured:
        console.print("[success]✓ AI authentication configured[/success]")
        if model_name:
            console.print(f"[success]✓ Model: {model_name}[/success]")
        console.print(
            "[warning]⚠ No database configured - you'll need to provide one via -d flag[/warning]"
        )

    console.print()
    console.print("[muted]Starting interactive session...[/muted]")
    console.print()


async def run_onboarding() -> bool:
    """Run the complete onboarding flow.

    Returns True if onboarding completed successfully (at least database configured),
    False if user cancelled or onboarding failed.
    """
    try:
        # Welcome screen
        welcome_screen()

        # Database setup
        database_name = await setup_database_guided()

        # If user cancelled database setup, exit
        if database_name is None:
            console.print("[warning]Database setup is required to continue.[/warning]")
            console.print(
                "[muted]You can also provide a connection string using: saber -d <connection-string>[/muted]"
            )
            return False

        # Auth setup
        auth_configured, model_name = await setup_auth_guided()

        # Show success screen
        success_screen(database_name, auth_configured, model_name)

        return True

    except KeyboardInterrupt:
        console.print("\n[warning]Onboarding cancelled.[/warning]")
        console.print(
            "[muted]You can run setup commands manually:[/muted]\n"
            "[muted]  - saber db add <name>  # Add database connection[/muted]\n"
            "[muted]  - saber auth setup     # Configure authentication[/muted]"
        )
        sys.exit(0)
    except Exception as e:
        console.print(f"[error]Onboarding failed:[/error] {e}")
        return False



import pytest

from sqlsaber import SQLSaber


@pytest.mark.asyncio
async def test_api_multiple_csvs_create_multiple_views(temp_dir, monkeypatch):
    config_dir = temp_dir / "config"
    monkeypatch.setattr(
        "platformdirs.user_config_dir", lambda *args, **kwargs: str(config_dir)
    )

    users = temp_dir / "users.csv"
    orders = temp_dir / "orders.csv"

    users.write_text("id,name\n1,Alice\n2,Bob\n", encoding="utf-8")
    orders.write_text("id,user_id,total\n10,1,9.99\n11,2,20.00\n", encoding="utf-8")

    saber = SQLSaber(
        database=[str(users), str(orders)],
        model_name="anthropic:claude-3-5-sonnet",
        api_key="test-key",
    )

    try:
        rows = await saber.connection.execute_query(
            'SELECT u.name, o.total FROM "users" u JOIN "orders" o ON u.id = o.user_id ORDER BY o.id'
        )
        assert rows == [
            {"name": "Alice", "total": 9.99},
            {"name": "Bob", "total": 20.0},
        ]
    finally:
        await saber.close()



"""Tests for the CSVsConnection backed by DuckDB."""

from urllib.parse import urlencode

import pytest

from sqlsaber.database import CSVsConnection
from sqlsaber.database.schema import SchemaManager


@pytest.mark.asyncio
async def test_csvs_connection_exposes_multiple_views(tmp_path):
    users = tmp_path / "users.csv"
    orders = tmp_path / "orders.csv"

    users.write_text("id,name\n1,Alice\n2,Bob\n", encoding="utf-8")
    orders.write_text("id,user_id,total\n10,1,9.99\n11,2,20.00\n", encoding="utf-8")

    specs = [f"csv:///{users}", f"csv:///{orders}"]
    conn = CSVsConnection(f"csvs:///?{urlencode({'spec': specs}, doseq=True)}")

    try:
        rows = await conn.execute_query(
            'SELECT u.name, o.total FROM "users" u JOIN "orders" o ON u.id = o.user_id ORDER BY o.id'
        )
        assert rows == [
            {"name": "Alice", "total": 9.99},
            {"name": "Bob", "total": 20.0},
        ]

        schema_manager = SchemaManager(conn)
        tables = await schema_manager.list_tables()
        names = {t["name"] for t in tables["tables"]}
        assert {"users", "orders"}.issubset(names)
    finally:
        await conn.close()


@pytest.mark.asyncio
async def test_csvs_connection_dedupes_view_names(tmp_path):
    dir_a = tmp_path / "a"
    dir_b = tmp_path / "b"
    dir_a.mkdir()
    dir_b.mkdir()

    csv_a = dir_a / "data.csv"
    csv_b = dir_b / "data.csv"

    csv_a.write_text("x\n1\n", encoding="utf-8")
    csv_b.write_text("x\n2\n", encoding="utf-8")

    specs = [f"csv:///{csv_a}", f"csv:///{csv_b}"]
    conn = CSVsConnection(f"csvs:///?{urlencode({'spec': specs}, doseq=True)}")

    try:
        rows = await conn.execute_query(
            'SELECT x FROM "data" UNION ALL SELECT x FROM "data_2" ORDER BY x'
        )
        assert rows == [{"x": 1}, {"x": 2}]

        schema_manager = SchemaManager(conn)
        tables = await schema_manager.list_tables()
        names = {t["name"] for t in tables["tables"]}
        assert {"data", "data_2"}.issubset(names)
    finally:
        await conn.close()
